###############################################################################
# Name: synglob.py                                                            #
# Purpose: Acts as a registration point for all supported languages.          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: synglob.py
AUTHOR: Cody Precord
@summary: Provides configuration and basic API functionality to all the syntax
          modules. It also acts  as a configuration file for the syntax
          management code. When support for a new languages is added it must
          have a registration entry in the below L{LANG_MAP} dictionary in
          order to be loadable by the syntax module.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: synglob.py 69887 2011-12-01 19:54:13Z CJP $"
__revision__ = "$Revision: 69887 $"

#-----------------------------------------------------------------------------#
# Dependencies
import wx.stc as stc

# The language identifiers and the EXT_MAP have been moved out of this
# module in order to be independent of Editra and wx, but they are
# still needed here...
from synextreg import *

#-----------------------------------------------------------------------------#
# Feature Identifiers
FEATURE_AUTOINDENT = u"AutoIndenter"
FEATURE_STYLETEXT = u"StyleText"

#-----------------------------------------------------------------------------#

# Maps file types to syntax definitions
LANG_MAP = {LANG_4GL    : (ID_LANG_4GL,       '_progress'),
            LANG_DSP56K : (ID_LANG_DSP56K,    '_asm68k'),
            LANG_68K    : (ID_LANG_68K,       '_asm68k'),
            LANG_ADA    : (ID_LANG_ADA,       '_ada'),
            LANG_APACHE : (ID_LANG_APACHE,    '_apache'),
            LANG_AS     : (ID_LANG_AS,        '_actionscript'),
            LANG_ASM    : (ID_LANG_ASM,       '_asm'),
            LANG_BASH   : (ID_LANG_BASH,      '_sh'),
            LANG_BATCH  : (ID_LANG_BATCH,     '_batch'),
            LANG_BOO    : (ID_LANG_BOO,       '_boo'),
            LANG_C      : (ID_LANG_C,         '_cpp'),
            LANG_CAML   : (ID_LANG_CAML,      '_caml'),
            LANG_CILK   : (ID_LANG_CILK,      '_cpp'),
            LANG_COBRA  : (ID_LANG_COBRA,     '_cobra'),
            LANG_COLDFUSION : (ID_LANG_COLDFUSION, '_html'),
            LANG_CPP    : (ID_LANG_CPP,    '_cpp'),
            LANG_CSH    : (ID_LANG_CSH,    '_sh'),
            LANG_CSHARP : (ID_LANG_CSHARP, '_cpp'),
            LANG_CSS    : (ID_LANG_CSS,    '_css'),
            LANG_D      : (ID_LANG_D,      '_d'),
            LANG_DIFF   : (ID_LANG_DIFF,   '_diff'),
            LANG_DJANGO : (ID_LANG_DJANGO, '_django'),
            LANG_DOT    : (ID_LANG_DOT,    '_dot'),
            LANG_EDJE   : (ID_LANG_EDJE,   '_edje'),
            LANG_EIFFEL : (ID_LANG_EIFFEL, '_eiffel'),
            LANG_ERLANG : (ID_LANG_ERLANG, '_erlang'),
            LANG_ESS    : (ID_LANG_ESS,    '_editra_ss'),
            LANG_F77    : (ID_LANG_F77,    '_fortran'),
            LANG_F95    : (ID_LANG_F95,    '_fortran'),
            LANG_FERITE : (ID_LANG_FERITE, '_ferite'),
            LANG_FLAGSHIP: (ID_LANG_FLAGSHIP, '_flagship'),
            LANG_FORTH  : (ID_LANG_FORTH, '_forth'),
            LANG_GLSL   : (ID_LANG_GLSL, '_glsl'),
            LANG_GUI4CLI : (ID_LANG_GUI4CLI,  '_gui4cli'),
            LANG_HASKELL : (ID_LANG_HASKELL,  '_haskell'),
            LANG_HAXE   : (ID_LANG_HAXE,      '_haxe'),
            LANG_HTML   : (ID_LANG_HTML,   '_html'),
            LANG_INNO   : (ID_LANG_INNO,   '_inno'),
            LANG_ISSL   : (ID_LANG_ISSL,   '_issuelist'),
            LANG_JAVA   : (ID_LANG_JAVA,   '_java'),
            LANG_JS     : (ID_LANG_JS,     '_javascript'),
            LANG_KIX    : (ID_LANG_KIX,    '_kix'),
            LANG_KSH    : (ID_LANG_KSH,    '_sh'),
            LANG_LATEX  : (ID_LANG_LATEX,  '_latex'),
            LANG_LISP   : (ID_LANG_LISP,   '_lisp'),
            LANG_LOUT   : (ID_LANG_LOUT,   '_lout'),
            LANG_LUA    : (ID_LANG_LUA,    '_lua'),
            LANG_MAKE   : (ID_LANG_MAKE,   '_make'),
            LANG_MAKO   : (ID_LANG_MAKO,    '_mako'),
            LANG_MASM   : (ID_LANG_MASM,   '_masm'),
            LANG_MATLAB : (ID_LANG_MATLAB, '_matlab'),
            LANG_MSSQL  : (ID_LANG_MSSQL,  '_mssql'),
            LANG_NASM   : (ID_LANG_NASM,   '_nasm'),
            LANG_NEWLISP: (ID_LANG_NEWLISP,'_lisp'),
            LANG_NONMEM : (ID_LANG_NONMEM, '_nonmem'),
            LANG_NSIS   : (ID_LANG_NSIS,   '_nsis'),
            LANG_OBJC   : (ID_LANG_OBJC,   '_cpp'),
            LANG_OCTAVE : (ID_LANG_OCTAVE, '_matlab'),
            LANG_OOC    : (ID_LANG_OOC,    '_ooc'),
            LANG_PASCAL : (ID_LANG_PASCAL, '_pascal'),
            LANG_PERL   : (ID_LANG_PERL,   '_perl'),
            LANG_PHP    : (ID_LANG_PHP,    '_php'),
            LANG_PIKE   : (ID_LANG_PIKE,   '_pike'),
            LANG_PLSQL  : (ID_LANG_PLSQL,  '_sql'),
            LANG_PROPS  : (ID_LANG_PROPS,  '_props'),
            LANG_PS     : (ID_LANG_PS,     '_postscript'),
            LANG_PYTHON : (ID_LANG_PYTHON, '_python'),
            LANG_R      : (ID_LANG_R,      '_s'),
            LANG_RUBY   : (ID_LANG_RUBY,   '_ruby'),
            LANG_S      : (ID_LANG_S,      '_s'),
            LANG_SCHEME : (ID_LANG_SCHEME, '_lisp'),
            LANG_SQL    : (ID_LANG_SQL,    '_sql'),
            LANG_SQUIRREL : (ID_LANG_SQUIRREL, '_squirrel'),
            LANG_ST     : (ID_LANG_ST,         '_smalltalk'),
            LANG_STATA : (ID_LANG_STATA,       '_stata'),
            LANG_SYSVERILOG : (ID_LANG_SYSVERILOG,  '_verilog'),
            LANG_TCL    : (ID_LANG_TCL,        '_tcl'),
            LANG_TXT    : (ID_LANG_TXT,        None),
            LANG_VALA   : (ID_LANG_VALA,       '_cpp'),
            LANG_VB     : (ID_LANG_VB,         '_visualbasic'),
            LANG_VBSCRIPT : (ID_LANG_VBSCRIPT, '_vbscript'),
            LANG_VERILOG: (ID_LANG_VERILOG,    '_verilog'),
            LANG_VHDL   : (ID_LANG_VHDL,    '_vhdl'),
            LANG_XML    : (ID_LANG_XML,     '_xml'),
            LANG_YAML   : (ID_LANG_YAML,    '_yaml'),
            LANG_GROOVY : (ID_LANG_GROOVY,  '_groovy'),
            LANG_XTEXT  : (ID_LANG_XTEXT,   '_xtext')
            }


### TODO: Profiling on the following methods to see if caching is necessary ###

# Dynamically finds the language description string that matches the given
# language id.
# Used when manually setting lexer from a menu/dialog
def GetDescriptionFromId(lang_id):
    """Get the programming languages description string from the given
    language id. If no corresponding language is found the plain text
    description is returned.
    @param lang_id: Language Identifier ID
    @note: requires that all languages are defined in ID_LANG_NAME, LANG_NAME
           pairs to work properly.

    """
    rval = LANG_TXT
    # Guard against async code that may be modifying globals
    globs = dict(globals())
    for key, val in globs.iteritems():
        if val == lang_id and key.startswith('ID_LANG'):
            rval = globs.get(key[3:], LANG_TXT)
            break
    return rval

def GetIdFromDescription(desc):
    """Get the language identifier for the given file type string. The search
    is case insensitive.
    @param desc: unicode (i.e u"Python")
    @note: if lookup fails ID_LANG_TXT is returned

    """
    rval = ID_LANG_TXT
    desc = desc.lower()
    # Guard against async code that may be modifying globals
    globs = dict(globals())
    for key, val in globs.iteritems():
        if isinstance(val, unicode):
            if val.lower() == desc and key.startswith('LANG_'):
                rval = globs.get("ID_" + key, ID_LANG_TXT)
                break
    return rval



###############################################################################
# Name: eclutil.py                                                            #
# Purpose: Common library utilities.                                          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Control Library: Editra Control Library Utility

Miscellaneous utility functions and gui helpers

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: eclutil.py 71178 2012-04-11 22:42:28Z CJP $"
__revision__ = "$Revision: 71178 $"

__all__ = ['AdjustAlpha', 'AdjustColour', 'BestLabelColour', 'HexToRGB',
           'GetHighlightColour', 'EmptyBitmapRGBA', 'Freezer',

           'DRAW_CIRCLE_SMALL', 'DRAW_CIRCLE_NORMAL', 'DRAW_CIRCLE_LARGE',
           'DrawCircleCloseBmp' ]

#-----------------------------------------------------------------------------#
# Imports
import wx

if wx.Platform == '__WXMAC__':
    try:
        import Carbon.Appearance
    except ImportError:
        CARBON = False
    else:
        CARBON = True

#-----------------------------------------------------------------------------#

# DrawCircleCloseBmp options
DRAW_CIRCLE_SMALL  = 0
DRAW_CIRCLE_NORMAL = 1
DRAW_CIRCLE_LARGE  = 2

__CircleDefs = { DRAW_CIRCLE_SMALL : dict(size=(8, 8),
                                          xpath=((1.75, 2), (4.75, 5),
                                                 (1.75, 5), (4.75, 2))),
                 DRAW_CIRCLE_NORMAL : dict(size=(16, 16),
                                           xpath=((4.5, 4), (10.5, 10),
                                                  (4.5, 10), (10.5, 4))),
                 DRAW_CIRCLE_LARGE : dict(size=(32, 32),
                                          xpath=((8, 8), (20, 20),
                                                 (8, 20), (20, 8))) }

#-----------------------------------------------------------------------------#
# Colour Utilities

def AdjustAlpha(colour, alpha):
    """Adjust the alpha of a given colour"""
    return wx.Colour(colour.Red(), colour.Green(), colour.Blue(), alpha)

def AdjustColour(color, percent, alpha=wx.ALPHA_OPAQUE):
    """ Brighten/Darken input colour by percent and adjust alpha
    channel if needed. Returns the modified color.
    @param color: color object to adjust
    @param percent: percent to adjust +(brighten) or -(darken)
    @keyword alpha: amount to adjust alpha channel

    """
    radj, gadj, badj = [ int(val * (abs(percent) / 100.0))
                         for val in color.Get() ]

    if percent < 0:
        radj, gadj, badj = [ val * -1 for val in [radj, gadj, badj] ]
    else:
        radj, gadj, badj = [ val or 255 for val in [radj, gadj, badj] ]

    red = min(color.Red() + radj, 255)
    green = min(color.Green() + gadj, 255)
    blue = min(color.Blue() + badj, 255)
    return wx.Colour(red, green, blue, alpha)

def BestLabelColour(color):
    """Get the best color to use for the label that will be drawn on
    top of the given color.
    @param color: background color that text will be drawn on

    """
    avg = sum(color.Get()) // 3
    if avg > 128:
        txt_color = wx.BLACK
    else:
        txt_color = wx.WHITE
    return txt_color

def GetHighlightColour():
    """Get the default highlight color
    @return: wx.Colour

    """
    if wx.Platform == '__WXMAC__':
        if CARBON:
            if hasattr(wx, 'MacThemeColour'):
                color = wx.MacThemeColour(Carbon.Appearance.kThemeBrushFocusHighlight)
                return color
            else:
                # kThemeBrushButtonPressedLightHighlight
                brush = wx.Brush(wx.BLACK)
                brush.MacSetTheme(Carbon.Appearance.kThemeBrushFocusHighlight)
                return brush.GetColour()

    # Fallback to text highlight color
    return wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)

def HexToRGB(hex_str):
    """Returns a list of red/green/blue values from a
    hex string.
    @param hex_str: hex string to convert to rgb

    """
    hexval = hex_str
    if hexval[0] == u"#":
        hexval = hexval[1:]
    ldiff = 6 - len(hexval)
    hexval += ldiff * u"0"
    # Convert hex values to integer
    red = int(hexval[0:2], 16)
    green = int(hexval[2:4], 16)
    blue = int(hexval[4:], 16)
    return [red, green, blue]

#-----------------------------------------------------------------------------#

def EmptyBitmapRGBA(width, height):
    """Create an empty bitmap with an alpha channel"""
    if hasattr(wx, 'EmptyBitmapRGBA'):
        bmp = wx.EmptyBitmapRGBA(width, height, alpha=0)
    else:
        bmp = wx.EmptyBitmap(width, height, -1)
        if hasattr(bmp, 'UseAlpha'):
            bmp.UseAlpha()
    return bmp

#-----------------------------------------------------------------------------#

class Freezer(object):
    """Context manager for freezing window redraws"""
    def __init__(self, window):
        super(Freezer, self).__init__()

        # Attributes
        self.window = window

    def __enter__(self):
        if self.window:
            self.window.Freeze()

    def __exit__( self, type, value, tb):
        if self.window:
            self.window.Thaw()

#-----------------------------------------------------------------------------#
# Drawing helpers

def DrawCircleCloseBmp(colour, backColour=None, option=DRAW_CIRCLE_SMALL):
    """
    Draws a small circular close button.
    @param colour: Circle's background colour
    @keyword backColour: pen colour for border and X
    @keyword option: DRAW_CIRCLE_* value
    @return: wxBitmap

    """
    assert option in __CircleDefs, "Invalid DRAW option!"

    defs = __CircleDefs.get(option)
    size = defs['size']
    if option != DRAW_CIRCLE_SMALL:
        # Adjust for border
        diameter = size[0] - 1
    else:
        diameter = size[0]
    radius = float(diameter) / 2.0
    xpath = defs['xpath']

    bmp = EmptyBitmapRGBA(size[0], size[1])
    dc = wx.MemoryDC()
    dc.SelectObject(bmp)
    dc.Clear()

    gc = wx.GraphicsContext.Create(dc)
    gc.SetBrush(wx.Brush(colour))
    if option > DRAW_CIRCLE_SMALL:
        gc.SetPen(wx.Pen(AdjustColour(colour, -30)))
    else:
        gc.SetPen(wx.TRANSPARENT_PEN)

    path = gc.CreatePath()
    path.AddCircle(radius, radius, radius)
    path.CloseSubpath()
    gc.FillPath(path)
    gc.StrokePath(path)

    path = gc.CreatePath()
    if backColour is not None:
        pen = wx.Pen(backColour, 1)
    else:
        pen = wx.Pen("white", 1)

    pen.SetCap(wx.CAP_BUTT)
    pen.SetJoin(wx.JOIN_BEVEL)
    gc.SetPen(pen)
    path.MoveToPoint(*xpath[0])
    path.AddLineToPoint(*xpath[1])
    path.MoveToPoint(*xpath[2])
    path.AddLineToPoint(*xpath[3])
    path.CloseSubpath()
    gc.DrawPath(path)

    dc.SelectObject(wx.NullBitmap)
    return bmp



###############################################################################
# Name: stata.py                                                              #
# Purpose: Define Stata syntax for highlighting and other features            #
# Author: Jean Eid<jeid@wlu.ca>                                               #
# Note: stata keywords are taken from TextWrangler module for stata           #
###############################################################################

"""
FILE: stata.py                                                                
@author: Jean Eid                                                      
@summary: Lexer configuration file for STATA source files.
                                                                         
"""

__author__ = "Jean Eid <jeid@wlu.ca>"
__svnid__ = "$Id: _stata.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 0$"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local imports
import synglob
import syndata
from _cpp import AutoIndenter

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

#  Documentation Keywords (includes Doxygen keywords)
SECONDARY_KEYWORDS =(1,
                     """
                     __GEEBT __GEERC  __GEEUC _3dax0 _3daxmin _3daxout if
                     _3daxtbl _3ddflts _3dmkdta _3dmnmx _3dproj _3drproj 
                     _3drshow _3dshad _3dsvusr _ac _addl _addop _adjksm 
                     _all _bsqreg _byobs _callerr _cpmatnm _cr1form 
                     _cr1invt _cr1se _cr1t _crc2use _crc4fld _crcacnt 
                     _crcar1 _crcause _crcbcrt _crcbin _crcbygr _crcchi2 
                     _crcchkt _crcchkw _crcci _crccip _crceprs _crcexn1 
                     _crcexn2 _crcexn4 _crcexn5 _crcexn6 _crcexn7 _crcexn8 
                     _crcexn9 _crcexna _crcexnb _crcexnc _crcexnd _crcexne 
                     _crcexnf _crcexnt _crcgldv _crcglil _crcichi _crcird 
                     _crcirr _crcksm _crclf _crcmeq _crcmiss _crcnlou 
                     _crcnms2 _crcnuse _crcor _crcphdr _crcplst _crcra 
                     _crcrd _crcrnfd _crcrr _crcrsfl _crcseq _crcshdr 
                     _crcslbl _crcsrv2 _crcsrvc _crcstep _crcswxx _crct 
                     _crctmge _crcunab _crcunit _crcvarl _crcwsrv _crczsku 
                     _cu_c0 _diparm _evlist _fracchk _fraccox _fracddp 
                     _fracdis _fracdv _fracin _fracmdp _fracord _fracpp 
                     _fracpv _fracrep _fracwgt _fracxo _gcount _gcut _gdiff 
                     _getbv _getrhs _getrres _gfill _ggroup _giqr _gladder 
                     _glmfl _glmilnk _glmmapf _glmmapl _glmresd _glmrpt 
                     _glmwgt _gma _gmad _gmax _gmdmean _gmdmed _gmean 
                     _gmedian _gmin _gmtr _gpctile _grank _grank2 _grfirst 
                     _grlast _grmax _grmean _grmin _grmiss _grmiss2 _grobs 
                     _grsd _grsum _gsd _gsrank _gstd _gsum _gtma _gtrank 
                     _hu _hub _hube _huber _inlist _invlist _isfit _jprfpdt 
                     _jprfpdx _jprfpfp _jprfpgn _jprfpin _jprfplx _jprfpmm 
                     _jprfppp _jprfpre _jprfprp _jprfpse _jprfptp _jprfpxo 
                     _jprglef _jprglfl _jprglil _jprglld _jprglwz _jprxrpa 
                     _kalman1 _ksmwrk _ldrtest _linemax _maked _merge 
                     _mfrmvec _mkvec _mvec _newey _nlout _nobs _opnum 
                     _parsevl _parsewt _partset _pctile _pred_me _pred_se 
                     _predict _qreg _repart _result _rmcoll _robksm _robust 
                     _sfran _subchar _svy _sw_lik _sw_lik2 _sw_ood _ts 
                     _ts_dsmp _ts_flag _ts_gdat _ts_meqn _ts_pars _ts_peri 
                     _tsheadr _ttest _ttest1 _ttest2 _tutends _tx_mtr1 
                     _tx_mtr2 _tx_mtr3 _tx_mtr4 _tx_mtr5 _tx_rpl _wkapm 
                     _wsrvcrv _xtrenorm
                     """)

MAIN_KEYWORDS =  (0, 
                  """
                  abbrev about abs acprplot add 
                  adjust ado adopath alpha an ano anov anova anovadef 
                  aorder ap app appe appen append arch arch_dr arch_p 
                  areg areg_p args arima arima_dr arima_p as ass asse 
                  asser assert at avplot avplots aw aweight bcskew0 be 
                  bee beep binreg bipp_lf bipr_lf bipr_p biprobit bitest 
                  bitesti bitowt blogit bmemsize boot bootsamp boxcox 
                  boxcox_p bprobit break brier bs bsampl_w bsample 
                  bsqreg bstat bstrap by bys bysort byte c_local canon 
                  canon_p capture cat cc cchart cci cd cell cellsize 
                  centile cf char char chdir checksum chi2 chi2tail ci 
                  cii cksum clear clo clocal clog clog_lf clog_p clogi 
                  clogi_sw clogit clogit_p clogitp clogl_sw cloglog 
                  close cluster cmdlog cnr cnre cnreg cnreg_sw cnsreg 
                  codebook col collaps4 collapse compare compress 
                  compute cond conf confi confir confirm cons const 
                  constr constra constrai constrain constraint contract 
                  copy cor corc corr corr2data corre correl correla 
                  correlat correlate corrgram cou coun count cox cox_p 
                  cox_sw coxbase coxhaz coxvar cp cprplot crc cross cs 
                  cscript csi ct ct_is ctset ctst_5 ctst_st cttost cumsp 
                  cumul cusum d datetof dbeta de debug debugbuf dec deco 
                  decod decode def deff define des desc descr descri 
                  describ describe dfbeta dfuller di dir dis discard 
                  disp disp_res disp_s displ displa display do doe doed 
                  doedi doedit dotplot dprobit drawnorm drop ds dstdize 
                  dwstat dydx dyex dynre dynren e ed edi edit egen 
                  eivreg else emdef en enc enco encod encode end eq 
                  eqlist erase ereg ereg_lf ereg_p ereg_sw err erro 
                  error est esti estim estima estimat estimate estimates 
                  etodow etof etomdy ex exact exec execu execut execute 
                  exi exit expand export eydx eyex F fac fact facto 
                  factor fast fft fillin findit fit float for for5_0 
                  force form forma format fpredict frac_154 frac_adj 
                  frac_chk frac_cox frac_ddp frac_dis frac_dv frac_in 
                  frac_mun frac_pp frac_pq frac_pv frac_wgt frac_xo 
                  fracgen fracplot fracpoly fracpred freq frequency 
                  Ftail ftodate ftoe ftomdy ftowdate fw fweight g gamma 
                  gamma_lf gamma_p gamma_sw ge gen gene gener genera 
                  generat generate genrank genstd genvmean gettoken gl 
                  gladder glm glm_p glm_sw glmpred glo glob globa global 
                  glogit glogit_p gnbre_lf gnbreg gnbreg_5 gnbreg_p 
                  gomp_lf gompe_sw gomper_p gompertz gph gphdot gphpen 
                  gphprint gprobi_p gprobit gr gr_print gra grap graph 
                  grebar greigen grmeanby group gsort gwood h hadimvo 
                  hareg hausman 
                  he heck_d2 heckma_p heckman heckp_lf heckpr_p heckprob 
                  hel help helpchk hereg hetpr_lf hetpr_p hetprob 
                  hettest hilite hist hlogit hlu hotel hprobit hreg icd9 
                  icd9p  iis impute in index inf infi infil infile 
                  infix inlist inp inpu input ins insh inshe inshee 
                  insheet insp inspe inspec inspect int integ intreg 
                  intrg_ll invchi2 invchi2tail invF invFtail invnchi2 
                  invnorm invttail ipolate iqreg ir iri istdize ivreg iw 
                  iweight joinby kalarma1 kap kap_3 kapmeier kappa 
                  kapwgt kdensity keep ksm ksmirnov ktau kwallis l la 
                  lab labe label ladder length level leverage lfit 
                  lfit_p li lincom linesize linktest lis list llogi_sw 
                  llogis_p llogist ln lnorm_lf lnorm_sw lnorma_p lnormal 
                  lnskew0 lo loc loca local log logi logis_lf logistic 
                  logit logit_p loglogs logrank logtype loneway long loo 
                  look lookfor looku lookup lower lpredict lroc lrtest 
                  ls lsens lsens_x lstat ltable ltrim lv lvr2plot m ma 
                  mac macr macro man mantel mark markout marksample mat 
                  matcell match matcol matcproc matname matr matri 
                  matrix matrow matsize matstrik max mcc mcci md0_ md1_ 
                  md1debu_ md2_ md2debu_ mdytoe mdytof mean means median 
                  memory memsize meqparse mer merg merge mfx mhodds min 
                  missing mkdir mkmat mkspline ml ml_5 ml_adjs ml_bhhhs 
                  ml_c_d ml_check ml_clear ml_cnt ml_debug ml_defd ml_e0 
                  ml_e0i ml_e1 ml_e2 ml_ebfg0 ml_ebfr0 ml_ebh0q ml_ebhh0 
                  ml_ebhr0 ml_ebr0i ml_ecr0i ml_edfp0 ml_edfr0 ml_edr0i 
                  ml_eds ml_eer0i ml_egr0i ml_elf ml_elfi ml_elfs 
                  ml_enr0i ml_enrr0 ml_exde ml_geqnr ml_grad0 ml_graph 
                  ml_hbhhh ml_hd0 ml_init ml_inv ml_log ml_max ml_mlout 
                  ml_model ml_nb0 ml_opt ml_plot ml_query ml_rdgrd 
                  ml_repor ml_s_e ml_searc mleval mlf_ mlmatsum mlog 
                  mlogi mlogit mlogit_p mlopts mlsum mlvecsum mnl0_ mor 
                  more mov move mrdu0_ mvdecode mvencode mvreg n nbreg 
                  nbreg_al nbreg_lf nbreg_sw nchi2 net newey newey_p 
                  news nl nl_p nlexp2 nlexp2a nlexp3 nlgom3 nlgom4 
                  nlinit nllog3 nllog4 nlogit nlpred no nobreak nod 
                  nodiscrete noe noesample nof nofreq noi nois noisi 
                  noisil noisily nol nolabel nonl nonlinear normden nose 
                  note notes notify now nowght npnchi2 nptrend numlist 
                  obs off old_ver olo olog ologi ologi_sw ologit 
                  ologit_p ologitp on one onew onewa oneway op_colnm 
                  op_comp op_diff op_inv op_str opr opro oprob oprob_sw 
                  oprobi oprobi_p oprobit oprobitp order orthog orthpoly 
                  ou out outf outfi outfil outfile outs outsh outshe 
                  outshee outsheet ovtest pac par pars parse pause pc 
                  pchart pchi pcorr pctile pentium percent pergram 
                  permanent personal pkcollapse pkcross pkequiv 
                  pkexamine pkshape pksumm pl playsnd plo plot plug 
                  plugi plugin pnorm poisgof poiss_lf poiss_sw poisso_p 
                  poisson pop popu popup post postclose postfile pperron 
                  pr prais prais_e prais_p pred predi predic predict 
                  predict preserve printgr priorest pro prob probi 
                  probit probit_p prog progr progra program prove prtest 
                  prtesti push pw pwcorr pwd pweight q qby qchi qnorm 
                  qqplot qreg qreg_c qreg_p qreg_sw qu quadchk quantile 
                  que quer query qui quie quiet quietl quietly range 
                  ranksum rawsum rchart rcof real recast recode reg reg3 
                  reg3_p regdw regr regre regre_p2 regres regres_p 
                  regress regriv_p remap ren rena renam rename renpfix 
                  repeat replace replay reshape restore ret retu retur 
                  return reverse rm roccomp rocfit rocgold roctab rot 
                  rota rotat rotate round row rreg rreg_p rtrim ru run 
                  runtest rvfplot rvpplot sa safesum sample sampsi sav 
                  save saving say sca scal scala scalar sco scob_lf 
                  scob_p scobi_sw scobit scor score sd sdtest sdtesti se 
                  search separate seperate serrbar set sfrancia sh she 
                  shel shell shewhart showpoint signrank signtest simul 
                  simulinit sktest sleep smcl smooth snapspan so sor 
                  sort spearman speedchk1 speekchk2 spikeplt spline_x 
                  sqreg sret sretu sretur sreturn st st_ct st_hc st_hcd 
                  st_is st_issys st_note st_promo st_set st_show st_smpl 
                  st_subid stack stackdepth stackreset statsby stbase 
                  stci stcox stcox_p stcoxkm stcurv stcurve stdes stem 
                  stereg stfill stgen stinit stir stjoin stmc stmh 
                  stphplot stphtest stptime strate streg streset string 
                  sts stset stsplit stsum sttocc sttoct stvary stweib su 
                  subinstr subinword subpop substr subwin sum summ summa 
                  summar summari summariz summarize sureg survcurv 
                  survsum svmat svy_disp svy_dreg svy_est svy_get 
                  svy_head svy_sub svy_x svydes svyintrg svyivreg svylc 
                  svylog_p svylogit svymean svymlog svyolog svyoprob 
                  svypois svyprobt svyprop svyratio svyreg svyreg_p 
                  svyset svytab svytest svytotal sw swcnreg swcox swereg 
                  swilk swlogis swlogit swologit swoprbt swpois swprobit 
                  swqreg swtobit swweib symmetry symmi symplot syntax 
                  sysdir sysmenu ta tab tab_or tab1 tab2 tabd tabdi 
                  tabdis tabdisp tabi table tabodds tabstat tabu tabul 
                  tabula tabulat tabulate te tempfile tempname tempvar 
                  tes test testnl testparm teststd text timer tis tob 
                  tobi tobit tobit_p tobit_sw token tokeni tokeniz 
                  tokenize touch treatreg trim truncreg tset tsfill 
                  tsreport tsrevar tsset tsunab ttail ttest ttesti 
                  tut_chk tut_wait tutorial ty typ type typeof u unab 
                  unabbrev uniform update upper us use using val values 
                  var variable varlabelpos vce verinst vers versi versio 
                  version vif vwls wdatetof wdctl wdlg wdupdate weib_lf 
                  weib_lf0 weibu_sw weibul_p weibull wh whelp whi whic 
                  which whil while wilc_st wilcoxon win wind windo 
                  window winexec winhelp wmenu wntestb wntestq xchart 
                  xcorr xi xpose xt_iis xt_tis xtabond xtbin_p xtclog 
                  xtcnt_p xtcorr xtdata xtdes xtgee xtgee_p xtgls 
                  xtgls_p xthaus xtile xtint_p xtintreg xtivreg xtlogit 
                  xtnb_fe xtnb_lf xtnbreg xtpcse xtpois xtpred xtprobit 
                  xtps_fe xtps_lf xtps_ren xtrch_p xtrchh xtrefe_p xtreg 
                  xtreg_be xtreg_fe xtreg_ml xtreg_re xtregar xtrere_p 
                  xtsum xttab xttest0 xttobit xttrans xwin xwind xwindo 
                  xwindow zap_s zinb zinb_llf zinb_plf zip zip_llf zip_p 
                  zip_plf zt_ct_5 zt_hc_5 zt_hcd_5 zt_is_5 zt_iss_5 
                  zt_sho_5 zt_smp_5 ztbase_5 ztcox_5 ztdes_5 ztereg_5 
                  ztfill_5 ztgen_5 ztir_5 ztjoin_5 zts_5 ztset_5 
                  ztspli_5 ztsum_5 zttoct_5 ztvary_5 ztweib_5 comma tab 
                  robust mvsktest oglm dlist mahapick viewresults 
                  valuesof textbarplot freduse hcavar distrate zipsave 
                  hutchens svylorenz goprobit regoprob digits cprplot2 
                  devcon oaxaca supclust convert_top_lines graphbinary 
                  trellis traces wtd glst lxpct_2 clv backrasch mmsrm 
                  diplot kountry gologit2 intcens svygei_svyatk fitint 
                  xtarsim eclpci beamplot samplepps classplot spearman2 
                  fixsort casefat mehetprob svyselmlog svybsamp2 
                  moremata checkfor2 dbmscopybatch palette_all ivreset 
                  lookfor_all nmissing isvar descogini akdensity 
                  biplotvlab genscore rrlogit ivvif hnblogit hnbclg 
                  hglogit hgclg hplogit hpclg ztg cdfplot kdens cpoisson 
                  samplesize gnbstrat nbstrat genass outreg2 kapprevi 
                  sampsi_reg lgamma2 glgamma2 ivgauss2 givgauss2 
                  sampsi_mcc cnbreg xtivreg2 sliceplot cquantile primes 
                  scheme_rbn1mono subsave insob geekel2d censornb 
                  estadd_plus estadd xtregre2 clorenz usmaps2 usswm 
                  tablemat tmap mif2dta surface kdens2 onespell diffpi 
                  plotbeta batplot relrank invcdf jmp jmp2 smithwelch 
                  reswage dfl p2ci heterogi shuffle8 shuffle lincheck 
                  bicdrop1 pre optifact matsort clustsens listmiss bic 
                  hdquantile dirlist variog svypxcon svypxcat seast 
                  xtlsdvc avplots4 eststo vclose gausshermite simirt 
                  cprplots glcurve sscsubmit sdecode harmby ellip 
                  mlboolean splitvallabels circular betacoef bnormpdf 
                  grqreg genhwcci mrtab zandrews cvxhull abar mat2txt 
                  gmlabvpos tableplot sortlistby mgen hlist ppplot rocss 
                  pnrcheck stexpect rbounds regaxis pgmhaz8 meta_lr 
                  mcqscore mvcorr ipf hapblock swblock raschtest 
                  gammasym raschtestv7 vanelteren fs estout1 estout 
                  loevh msp detect disjoint adjacent plotmatrix nnmatch 
                  senspec twoway_estfit mkdat soepren alphawgt center 
                  decompose wgttest cochran examples rollreg clemao_io 
                  heckprob2 gipf metagraph hshaz tslist collapse2 
                  gen_tail carryforward floattolong doubletofloat 
                  margeff matin4-matout4 perturb coldiag2 sim_arma ndbci 
                  labelmiss mcenter sslope reorder scat3 dummieslab rfl 
                  xtvc metareg rc2 moments sxpose kaputil bystore mice 
                  gzsave witch cureregr hprescott tabout gamet duncan 
                  fview eret2 rc_spline tolerance modeldiag metaparm 
                  profhap nsplit hlm fieller xtfisher matwrite usmaps 
                  ellip6 ellip7 xi3 qhapipf slist nearest fedit extremes 
                  mypkg pairplot cycleplot ciplot selectvars stcompet 
                  full_palette catplot eclplot spellutil metadialog 
                  psmatch2 bygap ingap mylabels metaaggr cleanlog gpfobl 
                  mvprobit eqprhistogram slideplot majority hireg bigtab 
                  vartyp codebook2 dmariano whotdeck crtest 
                  collapseunique stripplot linkplot statsbyfast parplot 
                  mitools groups wclogit xcontract xcollapse metafunnel 
                  corrtab dmerge makematrix cibplot vreverse msplot 
                  nicedates mkcorr nearmrg tabstatmat panelunit 
                  panelauto safedrop gammafit gumbelfit betafit 
                  lookforit savasas usesas etime usagelog tmpdir 
                  confirmdir shortdir lambda survtime xtabond2 
                  factortest checkvar vtokenize reshape8 pcorr2 tarow 
                  cb2html survwgt svr jonter xsampsi ci2 domdiag 
                  xtpattern nbfit distinct maketex triprobit smileplot 
                  tabmerge avplot3 datesum _gclsort cltest varlab decomp 
                  overlay tab3way tab2way ivreg2 hapipf varlag vececm 
                  isco isko ptrend dpplot cipolate tknz levels reformat 
                  xtab mvsumm fsum stkerhaz explist qlognorm tsspell 
                  gphepssj texteditors alphawgt decompose inccat adjksm 
                  outdat reshape7 qsim allpossible glmcorr gcause 
                  selmlog matsave est2tex log2do2 hansen2 gam ivhettest 
                  newey2 intext matrixof mrdum fastcd ivendog tabcount 
                  tabcond minap qrowname cij ciw textgph latab autolog 
                  histbox kdbox sunflower charlist adoedit lincomest 
                  stylerules strgen wntstmvq grnote xttrans2 inequal7 
                  tablepc hegy4 regdplot denormix chi2fit bstut aboutreg 
                  _gstd01 cpcorr mktab vecar xdatelist strdate thinplate 
                  gfields takelogs catgraph dsconcat tosql outseries 
                  glcurve7 omninorm summdate sencode rgroup cf3 hlpdir 
                  descsave outmat svytabs mstore savesome stbget spsurv 
                  xtgraph effects stpm madfuller aformat todate _gwtmean 
                  tsgraph soreg fbar metaninf blist johans vecar6 
                  outtable factmerg factref factext hadrilm levinlin 
                  nharvey ipshin gpreset geneigen dotex outtex sutex 
                  dsearch chiplot idonepsu statsmat ds3 dthaz paran 
                  gprefscode _gsoundex bpass bking labsort intreg2 sq 
                  powercal denton corr_svy log2html dfao xpredict mcl 
                  listtex raschcvt diagt estsave egenmore labutil 
                  concord avplot2 tablecol metabias coldiag fitstat 
                  doubmass cortesti fndmtch2 cusum6 ghistcum findval 
                  centcalc xrigls dfgls charutil icomp enlarge kpss 
                  metatrim ivgmm0 smhsiao matvsort roblpr modlpr recode2 
                  showgph copydesc shapley rnd himatrix bspline stcascoh 
                  stselpre nct ncf hist3 dolog dologx tscollap bcoeff 
                  grfreq grlogit lrchg lrmatx lrplot lrutil forfile 
                  printgph readlog spaces title dashln lomodrs ctabstat 
                  expandby finddup fitmacro normtest predcalc _grmedf 
                  ranvar matmap svmatf lincom2 csjl shownear fracdiff 
                  genvars calibr fracirf xriml rowranks tgraph ordplot 
                  cpr mlcoint stcmd xttest3 atkplot fsreg ciform rowsort 
                  expgen epsigr dashgph addtxt swboot stak _grprod 
                  sskapp xttest2 trinary ivprob-ivtobit6 torumm split q 
                  linesize keyb expr xtlist xtcount xtcorr2 vce2 summvl 
                  ststrata stcount simul2 regh pcamv pca mvpatt mokken 
                  lrtest2 linest iia hordered htest elogit diag wraplist 
                  qsort precmd modfycmd matfunc listuniq multgof hilite2 
                  unlabeld to_msp sample2 placevar listby listblck hh 
                  genl for2 dropvars countby bys icslib varcase tsplot 
                  diagtest ssizebi studysi poverty ineq dissim geivars 
                  ineqdec0 ineqdeco ineqfac povdeco sumdist xfrac 
                  dagumfit smfit cpyxplot mkbilogn univstat hotdeck
                  matodd p_vlists gennorm tab_chi sbplot5 sbplot
                  mfracpol keyplot taba ds5 tabplot cistat probitiv
                  tobitiv _gslope sphdist ginidesc inequal2 kernreg1
                  kernreg2 outfix outfix2 seg outreg rfregk spautoc
                  onewplot pwcorrs ltable2 tabhbar hbox tabhplot cihplot 
                  civplot sf36 pwploti partgam cf2 xtile2 ivglog 
                  kwallis2 far5 jb6 gby strparse _gprod mfilegr desmat 
                  areg2 margfx arimafit moreobs tsmktim durbinh bgtest 
                  mnthplot archlm gphudak renames skewplot cnsrsig 
                  recast2 doub2flt feldti tolower lfsum whitetst bpagan 
                  listutil mdensity kdmany stquant byvar cflpois 
                  workdays flower _grpos stcoxgof stcoxplt stpiece 
                  overid overidxt swapval adotype fndmtch svvarlbl 
                  gentrun storecmd sto lrdrop1 lrseq dmexogxt 
                  probexog-tobexog predxcon predxcat mmerge tablab 
                  for211 gmci grand nbinreg spikeplt ocratio biplot 
                  coranal mca heckman2 marker markov pgamma qgamma 
                  somersd pantest2 datmat distan missing quantil2 
                  distplot tpred contrast cid rglm dtobit2 ljs ewma 
                  barplot genfreq hbar hplot fodstr catdev rmanova 
                  ranova seq intterms lmoments regresby reglike pweibull 
                  wbull qweibull regpred logpred adjmean adjprop spell 
                  switchr trnbin0 mvsamp1i mvsampsi tpvar mvtest addtex 
                  pwcorrw vlist violin eba mstdize orthog stcumh 
                  ccweight psbayes oprobpr cndnmb3 pbeta qbeta vmatch 
                  kr20 sbrowni canon stbtcalc stgtcalc zb_qrm catenate 
                  lprplot nnest longplot parmest qqplot2 jb zip zinb 
                  hetprob unique longch gwhet williams adjust barplot2 
                  grand2 histplot xcorplot clarify mlogpred nproc 
                  labgraph vallist pexp qexp lms levene centroid medoid 
                  cluster fulltab t2way5 epiconf lstack deaton colelms 
                  confsvy median winsor bys torats venndiag chaos 
                  muxyplot muxplot irrepro triplot tomode circstat tryem 
                  white strip ralloc acplot stack symmetry omodel 
                  allcross dups iia sdtest vplplot summvl labsumm 
                  loopplot elapse istdize blogit2 sparl vallab gologit 
                  mkstrsn poisml trpois0 cenpois sssplot hausman stcstat
                  forvalues
                  """
)

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_C_DEFAULT, 'default_style'),
                 (stc.STC_C_COMMENT, 'comment_style'),
                 (stc.STC_C_COMMENTLINE, 'comment_style'),
                 (stc.STC_C_COMMENTDOC, 'comment_style'),
                 (stc.STC_C_COMMENTDOCKEYWORD, 'dockey_style'),
                 (stc.STC_C_COMMENTDOCKEYWORDERROR, 'error_style'),
                 (stc.STC_C_COMMENTLINE, 'comment_style'),
                 (stc.STC_C_COMMENTLINEDOC, 'comment_style'),
                 (stc.STC_C_CHARACTER, 'char_style'),
                 (stc.STC_C_GLOBALCLASS, 'global_style'),
                 (stc.STC_C_IDENTIFIER, 'default_style'),
                 (stc.STC_C_NUMBER, 'number_style'),
                 (stc.STC_C_OPERATOR, 'operator_style'),
                 (stc.STC_C_PREPROCESSOR, 'pre_style'),
                 (stc.STC_C_REGEX, 'pre_style'),
                 (stc.STC_C_STRING, 'string_style'),
                 (stc.STC_C_STRINGEOL, 'default_style'),
                 (stc.STC_C_UUID, 'pre_style'),
                 (stc.STC_C_VERBATIM, 'number2_style'),
                 (stc.STC_C_WORD, 'keyword_style'),
                 (stc.STC_C_WORD2, 'keyword2_style') ]

#---- Extra Properties ----#
FOLD = ("fold", "1")
FOLD_PRE = ("styling.within.preprocessor", "0")
FOLD_COM = ("fold.comment", "1")
FOLD_COMP = ("fold.compact", "1")
FOLD_ELSE = ("fold.at.else", "0")

#------------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for STATA""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_CPP)
        self.RegisterFeature(synglob.FEATURE_AUTOINDENT, AutoIndenter)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [MAIN_KEYWORDS, SECONDARY_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD, FOLD_PRE, FOLD_COM]

    # TODO: this doesnt' look right...
    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'//', u'/*', u'*/',u'*' ]



###############################################################################
# Name: choicedlg.py                                                          #
# Purpose: Generic Choice Dialog                                              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: Choice Dialog

A generic choice dialog that uses a wx.Choice control to display its choices.

@summary: Generic Choice Dialog

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: choicedlg.py 70230 2012-01-01 01:47:42Z CJP $"
__revision__ = "$Revision: 70230 $"

__all__ = ['ChoiceDialog',]

#--------------------------------------------------------------------------#
# Imports
import wx

import ecbasewin

#--------------------------------------------------------------------------#
# Globals
ChoiceDialogNameStr = u"ChoiceDialog"

#--------------------------------------------------------------------------#

class ChoiceDialog(ecbasewin.ECBaseDlg):
    """Dialog with a wx.Choice control for showing a list of choices"""
    def __init__(self, parent, id=wx.ID_ANY,
                 msg=u'', title=u'',
                 choices=None, default=u'',
                 pos=wx.DefaultPosition,
                 size=wx.DefaultSize,
                 style=0,
                 name=ChoiceDialogNameStr):
        """Create the choice dialog
        @param parent: Parent Window
        @keyword id: Dialog ID
        @keyword msg: Dialog Message
        @keyword title: Dialog Title
        @keyword choices: list of strings
        @keyword default: Default selection
        @keyword pos: Dialog Position
        @keyword size: Dialog Size
        @keyword style: Dialog Style bitmask
        @keyword name: Dialog Name

        """
        super(ChoiceDialog, self).__init__(parent, id, title,
                                           style=wx.CAPTION, pos=pos,
                                           size=size, name=name)

        # Attributes
        panel = ChoicePanel(self, msg=msg,
                                  choices=choices,
                                  default=default,
                                  style=style)
        self.SetPanel(panel)
        self.SetInitialSize()

#--------------------------------------------------------------------------#

class ChoicePanel(wx.Panel):
    """Generic Choice dialog panel"""
    def __init__(self, parent, msg=u'', choices=list(),
                 default=u'', style=wx.OK|wx.CANCEL):
        """Create the panel
        @param parent: Parent Window
        @keyword msg: Display message
        @keyword choices: list of strings
        @keyword default: default selection
        @keyword style: panel style

        """
        super(ChoicePanel, self).__init__(parent)

        # Attributes
        self._msg = msg
        self._choices = wx.Choice(self, wx.ID_ANY)
        self._selection = default
        self._selidx = 0
        self._bmp = None
        self._buttons = list()

        # Setup
        self._choices.SetItems(choices)
        if default in choices:
            self._choices.SetStringSelection(default)
            self._selidx = self._choices.GetSelection()
        else:
            self._choices.SetSelection(0)
            self._selidx = 0
            self._selection = self._choices.GetStringSelection()

        # Setup Buttons
        for btn, id_ in ((wx.OK, wx.ID_OK), (wx.CANCEL, wx.ID_CANCEL),
                         (wx.YES, wx.ID_YES), (wx.NO, wx.ID_NO)):
            if btn & style:
                button = wx.Button(self, id_)
                self._buttons.append(button)

        if not len(self._buttons):
            self._buttons.append(wx.Button(self, wx.ID_OK))
            self._buttons.append(wx.Button(self, wx.ID_CANCEL))

        # Layout
        self.__DoLayout(style)

        # Event Handlers
        self.Bind(wx.EVT_CHOICE, self.OnChoice, self._choices)
        self.Bind(wx.EVT_BUTTON, self.OnButton)

    def __DoLayout(self, style):
        """Layout the panel"""
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        vsizer = wx.BoxSizer(wx.VERTICAL)
        caption = wx.StaticText(self, label=self._msg)

        # Layout the buttons
        bsizer = wx.StdDialogButtonSizer()
        for button in self._buttons:
            bsizer.AddButton(button)
            bid = button.GetId()
            if bid in (wx.ID_NO, wx.ID_YES):
                if wx.NO_DEFAULT & style:
                    if bid == wx.ID_NO:
                        button.SetDefault()
                else:
                    if bid == wx.ID_YES:
                        button.SetDefault()
            elif bid == wx.ID_OK:
                button.SetDefault()

        bsizer.Realize()

        vsizer.AddMany([((10, 10), 0), (caption, 0), ((20, 20), 0),
                        (self._choices, 1, wx.EXPAND), ((10, 10), 0),
                        (bsizer, 1, wx.EXPAND),
                        ((10, 10), 0)])

        icon_id = wx.ART_INFORMATION
        for i_id, a_id in ((wx.ICON_ERROR, wx.ART_ERROR),
                     (wx.ICON_WARNING, wx.ART_WARNING)):
            if i_id & style:
                icon_id = a_id
                break

        icon = wx.ArtProvider.GetBitmap(icon_id, wx.ART_MESSAGE_BOX, (64, 64))
        self._bmp = wx.StaticBitmap(self, bitmap=icon)
        bmpsz = wx.BoxSizer(wx.VERTICAL)
        bmpsz.AddMany([((10, 10), 0), (self._bmp, 0, wx.ALIGN_CENTER_VERTICAL),
                       ((10, 30), 0, wx.EXPAND)])
        hsizer.AddMany([((10, 10), 0), (bmpsz, 0, wx.ALIGN_TOP),
                        ((10, 10), 0), (vsizer, 1), ((10, 10), 0)])

        self.SetSizer(hsizer)
        self.SetInitialSize()
        self.SetAutoLayout(True)

    def GetChoiceControl(self):
        """Get the dialogs choice control
        @return: wx.Choice

        """
        return self._choices

    @ecbasewin.expose(ChoiceDialog)
    def GetSelection(self):
        """Get the chosen index
        @return: int

        """
        return self._selidx

    @ecbasewin.expose(ChoiceDialog)
    def GetStringSelection(self):
        """Get the chosen string
        @return: string

        """
        return self._selection

    def OnButton(self, evt):
        """Handle button events
        @param evt: wx.EVT_BUTTON

        """
        self.GetParent().EndModal(evt.GetId())

    def OnChoice(self, evt):
        """Update the selection
        @param evt: wx.EVT_CHOICE

        """
        if evt.GetEventObject() == self._choices:
            self._selection = self._choices.GetStringSelection()
            self._selidx = self._choices.GetSelection()
        else:
            evt.Skip()

    @ecbasewin.expose(ChoiceDialog)
    def SetBitmap(self, bmp):
        """Set the dialogs bitmap
        @param bmp: wx.Bitmap

        """
        self._bmp.SetBitmap(bmp)
        self.Layout()

    @ecbasewin.expose(ChoiceDialog)
    def SetChoices(self, choices):
        """Set the dialogs choices
        @param choices: list of strings

        """
        self._choices.SetItems(choices)
        self._choices.SetSelection(0)
        self._selection = self._choices.GetStringSelection()

    @ecbasewin.expose(ChoiceDialog)
    def SetSelection(self, sel):
        """Set the selected choice
        @param sel: int

        """
        self._choices.SetSelection(sel)
        self._selection = self._choices.GetStringSelection()
        self._selidx = self._choices.GetSelection()

    @ecbasewin.expose(ChoiceDialog)
    def SetStringSelection(self, sel):
        """Set the selected choice
        @param sel: string

        """
        self._choices.SetStringSelection(sel)
        self._selection = self._choices.GetStringSelection()
        self._selidx = self._choices.GetSelection()

#--------------------------------------------------------------------------#



###############################################################################
# Name: flagship.py                                                           #
# Purpose: Define Flagship syntax for highlighting and other features         #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: flagship.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for the Flagship programming language and
          other XBase dialects.
@todo: Custom style defs

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _flagship.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
FS_COMMANDS = (0, "? @ accept access all alternate announce ansi any append as "
                  "assign autolock average begin bell bitmap blank box call "
                  "cancel case century charset checkbox clear close cls color "
                  "combobox commit confirm console constant continue copy "
                  "count create cursor date dbread dbwrite decimals declare "
                  "default delete deleted delimiters device dir directory "
                  "display do draw edit else elseif eject end endcase enddo "
                  "endif endtext epoch erase error escape eval eventmask exact "
                  "exclusive extended external extra field file filter find "
                  "fixed font for form format from get gets global "
                  "global_extern go goto gotop guialign guicolor guicursor "
                  "guitransl html htmltext if image index input intensity join "
                  "key keyboard keytransl label lines list listbox local "
                  "locate margin memory memvar menu message method multibyte "
                  "multilocks next nfs nfslock nfs_force note on openerror "
                  "order outmode pack parameters path pixel pop printer "
                  "private prompt public push pushbutton quit radiobutton "
                  "radiogroup read recall refresh reindex relation release "
                  "rename replace report request restore richtext rowadapt "
                  "rowalign run save say scoreboard scrcompress screen seek "
                  "select sequence set setenhanced setstandard setunselected "
                  "skip softseek sort source static store struct structure sum "
                  "tag tbrowse text to total type typeahead unique unlock "
                  "update use wait while with wrap xml zap zerobyteout")

FS_STDLIB = (1, "_displarr _displarrerr _displarrstd _displobj _displobjerr "
                "_displobjstd aadd abs achoice aclone acopy adel adir "
                "aelemtype aeval afields afill ains alert alias alltrim altd "
                "ansi2oem appiomode appmdimode appobject array asc ascan asize "
                "asort at atail atanychar autoxlock between bin2i bin2l bin2w "
                "binand binlshift binor binrshift binxor bof break browse cdow "
                "chr chr2screen cmonth col col2pixel color2rgb colorselect "
                "colvisible consoleopen consolesize crc32 ctod curdir date "
                "datevalid day dbappend dbclearfilter dbclearindex "
                "dbclearrelation dbcloseall dbclosearea dbcommit dbcommitall "
                "dbcreate dbcreateindex dbdelete dbedit dbeval dbf dbfilter "
                "dbfinfo dbflock dbfused dbgetlocate dbgobottom dbgoto dbgotop "
                "dbobject dbrecall dbreindex dbrelation dbrlock dbrlocklist "
                "dbrselect dbrunlock dbseek dbselectarea dbsetdriver "
                "dbsetfilter dbsetindex dbsetlocate dbsetorder dbsetrelation "
                "dbskip dbstruct dbunlock dbunlockall dbusearea default "
                "deleted descend devout devoutpict devpos directory diskspace "
                "dispbegin dispbox dispcount dispend dispout doserror "
                "doserror2str dow drawline dtoc dtos empty eof errorblock "
                "errorlevel eval execname execpidnum exp fattrib fclose fcount "
                "fcreate ferase ferror ferror2str fieldblock fielddeci "
                "fieldget fieldgetarr fieldlen fieldname fieldpos fieldput "
                "fieldputarr fieldtype fieldwblock file findexefile fklabel "
                "fkmax flagship_dir flock flockf fopen found fread freadstdin "
                "freadstr freadtxt frename fs_set fseek fwrite getactive "
                "getalign getapplykey getdosetkey getenv getenvarr getfunction "
                "getpostvalid getprevalid getreader guidrawline hardcr header "
                "hex2num i2bin iif indexcheck indexcount indexdbf indexext "
                "indexkey indexnames indexord infobox inkey inkey2read "
                "inkey2str inkeytrap instdchar instdstring int int2num isalpha "
                "isbegseq iscolor isdbexcl isdbflock isdbmultip isdbmultiple "
                "isdbmultipleopen isdbrlock isdigit isfunction isguimode "
                "islower isobjclass isobjequiv isobjproperty isprinter isupper "
                "l2bin lastkey lastrec left len listbox lock log lower ltrim "
                "lupdate macroeval macrosubst max max_col max_row maxcol "
                "maxrow mcol mdblck mdiclose mdiopen mdiselect memocode "
                "memodecode memoedit memoencode memoline memoread memory "
                "memotran memowrit memvarblock mhide min minmax mlcount "
                "mlctopos mleftdown mlpos mod month mpostolc mpresent "
                "mreststate mrightdown mrow msavestate msetcursor msetpos "
                "mshow mstate neterr netname nextkey num2hex num2int objclone "
                "oem2ansi onkey ordbagext ordbagname ordcond ordcondset "
                "ordcreate orddescend orddestroy ordfor ordisinique ordkey "
                "ordkeyadd ordkeycount ordkeydel ordkeygoto ordkeyno ordkeyval "
                "ordlistadd ordlistclear ordlistrebui ordname ordnumber "
                "ordscope ordsetfocu ordsetrelat ordskipunique os outerr "
                "outstd padc padl padr param parameters pcalls pcol pcount "
                "pixel2col pixel2row printstatus procfile procline procname "
                "procstack proper prow qout qout2 qqout qqout2 rat rddlist "
                "rddname rddsetdefault readexit readinsert readkey readkill "
                "readmodal readsave readupdated readvar reccount recno recsize "
                "replicate restscreen right rlock rlockverify round row "
                "row2pixel rowadapt rowvisible rtrim savescreen scrdos2unix "
                "screen2chr scroll scrunix2dos seconds secondscpu select "
                "serial set setansi setblink setcancel setcol2get setcolor "
                "setcolorba setcursor setevent setguicursor setkey setmode "
                "setpos setprc setvarempty sleep sleepms soundex space sqrt "
                "statbarmsg statusmessage stod str strlen strlen2col "
                "strlen2pix strlen2space strpeek strpoke strtran strzero stuff "
                "substr tbcolumnnew tbmouse tbrowsearr tbrowsedb tbrowsenew "
                "tempfilename time tone transform trim truepath type updated "
                "upper used usersactive usersdbf usersmax val valtype version "
                "webdate weberrorhandler webgetenvir webgetformdata "
                "webhtmlbegin webhtmlend weblogerr webmaildomain weboutdata "
                "websendmail word year")

FS_FUNC = (2, "function procedure return exit")

FS_CLASS = (3, "class instance export hidden protect prototype")
#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_FS_ASM, ''),
                (stc.STC_FS_BINNUMBER, 'number_style'),
                (stc.STC_FS_COMMENT, 'comment_style'),
                (stc.STC_FS_COMMENTDOC, 'dockey_style'),
                (stc.STC_FS_COMMENTDOCKEYWORD, 'dockey_style'),
                (stc.STC_FS_COMMENTDOCKEYWORDERROR, 'error_style'),
                (stc.STC_FS_COMMENTLINE, 'comment_style'),
                (stc.STC_FS_COMMENTLINEDOC, 'comment_style'),
                (stc.STC_FS_CONSTANT, 'default_style'),
                (stc.STC_FS_DATE, 'default_style'),
                (stc.STC_FS_DEFAULT, 'default_style'),
                (stc.STC_FS_ERROR, 'error_style'),
                (stc.STC_FS_HEXNUMBER, 'number_style'),
                (stc.STC_FS_IDENTIFIER, 'default_style'),
                (stc.STC_FS_KEYWORD, 'keyword_style'),
                (stc.STC_FS_KEYWORD2, 'keyword2_style'),
                (stc.STC_FS_KEYWORD3, 'keyword3_style'),
                (stc.STC_FS_KEYWORD4, 'keyword4_style'),
                (stc.STC_FS_LABEL, 'default_style'),
                (stc.STC_FS_NUMBER, 'number_style'),
                (stc.STC_FS_OPERATOR, 'operator_style'),
                (stc.STC_FS_PREPROCESSOR, 'pre_style'),
                (stc.STC_FS_STRING, 'string_style'),
                (stc.STC_FS_STRINGEOL, 'stringeol_style')]

#---- Extra Properties ----#
FOLD = ('fold', '1')

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Flagship""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_FLAGSHIP)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [FS_COMMANDS, FS_STDLIB, FS_FUNC, FS_CLASS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'//']



###############################################################################
# Name: errdlg.py                                                             #
# Purpose: Error Reporter Dialog                                              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: Error Reporter Dialog

Dialog for displaying exceptions and reporting errors to application maintainer.
This dialog is intended as a base class and should be subclassed to fit the
applications needs.

This dialog should be initiated inside of a sys.excepthook handler.

Example:

sys.excepthook = ExceptHook
...
def ExceptionHook(exctype, value, trace):
    # Format the traceback
    ftrace = ErrorDialog.FormatTrace(exctype, value, trace)

    # Ensure that error gets raised to console as well
    print ftrace

    # If abort has been set and we get here again do a more forceful shutdown
    if ErrorDialog.ABORT:
        os._exit(1)

    # Prevent multiple reporter dialogs from opening at once
    if not ErrorDialog.REPORTER_ACTIVE and not ErrorDialog.ABORT:
        dlg = ErrorDialog(ftrace)
        dlg.ShowModal()
        dlg.Destroy()

@summary: Error Reporter Dialog

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: errdlg.py 70230 2012-01-01 01:47:42Z CJP $"
__revision__ = "$Revision: 70230 $"

__all__ = [# Classes
           'ErrorDialog', 'ErrorReporter',
           # Functions
           'TimeStamp']

#----------------------------------------------------------------------------#
# Dependencies
import os
import sys
import platform
import time
import traceback
import wx

# Local Imports
import ecbasewin

#----------------------------------------------------------------------------#
# Globals
_ = wx.GetTranslation

#----------------------------------------------------------------------------#

class ErrorReporter(object):
    """Crash/Error Reporter Service
    @summary: Stores all errors caught during the current session.
    @note: singleton class

    """
    instance = None
    _first = True
    def __init__(self):
        """Initialize the reporter
        @note: The ErrorReporter is a singleton.

        """
        # Ensure init only happens once
        if self._first:
            super(ErrorReporter, self).__init__()
            self._first = False
            self._sessionerr = list()
        else:
            pass

    def __new__(cls, *args, **kargs):
        """Maintain only a single instance of this object
        @return: instance of this class

        """
        if not cls.instance:
            cls.instance = object.__new__(cls, *args, **kargs)
        return cls.instance

    def AddMessage(self, msg):
        """Adds a message to the reporters list of session errors
        @param msg: The Error Message to save

        """
        if msg not in self._sessionerr:
            self._sessionerr.append(msg)

    def GetErrorStack(self):
        """Returns all the errors caught during this session
        @return: formatted log message of errors

        """
        return (os.linesep * 2).join(self._sessionerr)

    def GetLastError(self):
        """Gets the last error from the current session
        @return: Error Message String

        """
        if len(self._sessionerr):
            return self._sessionerr[-1]
        
#-----------------------------------------------------------------------------#

class ErrorDialog(ecbasewin.ECBaseDlg):
    """Dialog for showing errors and and notifying Editra.org should the
    user choose so.

    """
    ID_SEND = wx.NewId()
    ABORT = False
    REPORTER_ACTIVE = False
    def __init__(self, parent, id=wx.ID_ANY, title=u'',
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                 name="ErrorReporterDlg", message=u''):
        """Initialize the dialog
        @param parent: Parent Window
        @keyword id: Dialog ID
        @keyword title: Dialog Title
        @keyword pos: Dialog Postion
        @keyword size: Dialog Size
        @keyword style: Dialog Style bitmask
        @keyword name: Dialog Name
        @param message: Error message to display

        """
        ErrorDialog.REPORTER_ACTIVE = True
        super(ErrorDialog, self).__init__(parent, id, title, pos,
                                          size, style, name)
        
        # Give message to ErrorReporter
        ErrorReporter().AddMessage(message)

        # Attributes
        self.err_msg = os.linesep.join((self.GetEnvironmentInfo(),
                                        "#---- Traceback Info ----#",
                                        ErrorReporter().GetErrorStack(),
                                        "#---- End Traceback Info ----#"))

        # Layout
        self.SetPanel(ErrorPanel(self, self.err_msg))
        self.SetMinSize(wx.Size(450, 300))

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnButton)
        self.Bind(wx.EVT_CLOSE, self.OnClose)

        # Auto show at end of init
        self.CenterOnParent()

    #---- Override in Subclass ----#

    def Abort(self):
        """Called to abort the application
        @note: needs to be overridden in subclasses

        """
        raise NotImplementedError("Abort must be implemented!")

    def GetEnvironmentInfo(self):
        """Get the environmental info / Header of error report
        @return: string

        """
        info = list()
        info.append("#---- Notes ----#")
        info.append("Please provide additional information about the crash here")
        info.extend(["", ""])
        info.append("#---- System Information ----#")
        info.append(self.GetProgramName())
        info.append("Operating System: %s" % wx.GetOsDescription())
        if sys.platform == 'darwin':
            info.append("Mac OSX: %s" % platform.mac_ver()[0])
        info.append("Python Version: %s" % sys.version)
        info.append("wxPython Version: %s" % wx.version())
        info.append("wxPython Info: (%s)" % ", ".join(wx.PlatformInfo))
        info.append("Python Encoding: Default=%s  File=%s" % \
                    (sys.getdefaultencoding(), sys.getfilesystemencoding()))
        info.append("wxPython Encoding: %s" % wx.GetDefaultPyEncoding())
        info.append("System Architecture: %s %s" % (platform.architecture()[0], \
                                                    platform.machine()))
        info.append("Byte order: %s" % sys.byteorder)
        info.append("Frozen: %s" % str(getattr(sys, 'frozen', 'False')))
        info.append("#---- End System Information ----#")
        info.append("")
        return os.linesep.join(info)

    def GetProgramName(self):
        """Get the program name/version info to include in error report
        @return: string

        """
        return wx.GetApp().GetAppName()

    def Send(self):
        """Called to send error report
        @note: needs to be overridden in subclasses

        """
        raise NotImplementedError("Send must be implemented!")

    #---- End Required overrides ----#

    @staticmethod
    def FormatTrace(exctype, value, trace):
        """Format the traceback
        @return: string

        """
        exc = traceback.format_exception(exctype, value, trace)
        exc.insert(0, u"*** %s ***%s" % (TimeStamp(), os.linesep))
        ftrace = u"".join(exc)
        return ftrace

    def SetDescriptionLabel(self, label):
        """Set the dialogs main description text
        @param label: string

        """
        self._panel.SetDescriptionText(label)

    def ShowAbortButton(self, show=True):
        """Show/Hide the Abort button
        @keyword show: bool

        """
        btn = self._panel.FindWindowById(wx.ID_ABORT)
        if btn is not None:
            btn.Show(show)
            self._panel.Layout()

    def ShowSendButton(self, show=True):
        """Show/Hide the Send button
        @keyword show: bool

        """
        btn = self._panel.FindWindowById(ErrorDialog.ID_SEND)
        if btn is not None:
            btn.Show(show)
            self._panel.Layout()

    #---- Event Handlers ----#

    def OnButton(self, evt):
        """Handles button events
        @param evt: event that called this handler
        @postcondition: Dialog is closed
        @postcondition: If Report Event then email program is opened

        """
        e_id = evt.GetId()
        if e_id == wx.ID_CLOSE:
            self.Close()
        elif e_id == ErrorDialog.ID_SEND:
            self.Send()
            self.Close()
        elif e_id == wx.ID_ABORT:
            ErrorDialog.ABORT = True
            self.Abort()
            self.Close()
        else:
            evt.Skip()

    def OnClose(self, evt):
        """Cleans up the dialog when it is closed
        @param evt: Event that called this handler

        """
        ErrorDialog.REPORTER_ACTIVE = False
        evt.Skip()

#-----------------------------------------------------------------------------#

class ErrorPanel(wx.Panel):
    """Error Reporter panel"""
    def __init__(self, parent, msg):
        """Create the panel
        @param parent: wx.Window
        @param msg: Error message to display

        """
        super(ErrorPanel, self).__init__(parent)

        # Attributes
        self.err_msg = msg
        self.desc = wx.StaticText(self, label=u'')

        # Layout
        self.__DoLayout()

    def __DoLayout(self):
        """Layout the control"""
        icon = wx.StaticBitmap(self, 
                               bitmap=wx.ArtProvider.GetBitmap(wx.ART_ERROR))
        t_lbl = wx.StaticText(self, label=_("Error Traceback:"))
        tctrl = wx.TextCtrl(self, value=self.err_msg, style=wx.TE_MULTILINE | 
                                                            wx.TE_READONLY)

        abort_b = wx.Button(self, wx.ID_ABORT, _("Abort"))
        abort_b.SetToolTipString(_("Exit the application"))
        send_b = wx.Button(self, ErrorDialog.ID_SEND, _("Report Error"))
        send_b.SetDefault()
        close_b = wx.Button(self, wx.ID_CLOSE)

        # Layout
        vsizer = wx.BoxSizer(wx.VERTICAL)

        hsizer1 = wx.BoxSizer(wx.HORIZONTAL)
        hsizer1.AddMany([((5, 5), 0), (icon, 0, wx.ALIGN_CENTER_VERTICAL),
                         ((12, 5), 0), (self.desc, 0, wx.EXPAND), ((5, 5), 0)])

        hsizer2 = wx.BoxSizer(wx.HORIZONTAL)
        hsizer2.AddMany([((5, 5), 0), (tctrl, 1, wx.EXPAND), ((5, 5), 0)])

        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        bsizer.AddMany([((5, 5), 0), (abort_b, 0), ((-1, -1), 1, wx.EXPAND),
                        (send_b, 0), ((5, 5), 0), (close_b, 0), ((5, 5), 0)])

        vsizer.AddMany([((5, 5), 0),
                        (hsizer1, 0),
                        ((10, 10), 0),
                        (t_lbl, 0, wx.ALIGN_LEFT|wx.LEFT, 5),
                        ((3, 3), 0),
                        (hsizer2, 1, wx.EXPAND),
                        ((8, 8), 0),
                        (bsizer, 0, wx.EXPAND),
                        ((8, 8), 0)])

        self.SetSizer(vsizer)
        self.SetAutoLayout(True)

    def SetDescriptionText(self, text):
        """Set the description label text
        @param text: string

        """
        self.desc.SetLabel(text)
        self.Layout()

#-----------------------------------------------------------------------------#

def TimeStamp():
    """Create a formatted time stamp of current time
    @return: Time stamp of the current time (Day Month Date HH:MM:SS Year)

    """
    now = time.localtime(time.time())
    now = time.asctime(now)
    return now



###############################################################################
# Name: testSyntaxDataBase.py                                                 #
# Purpose: Unit tests for synextreg.ExtensionRegister                         #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2010 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing synextreg.ExtensionRegister class"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testSyntaxExtReg.py 63844 2010-04-03 17:31:54Z CJP $"
__revision__ = "$Revision: 63844 $"

#-----------------------------------------------------------------------------#
# Imports
import unittest
import os
import wx
import wx.stc

# Module to test
import syntax.synextreg as synextreg

#-----------------------------------------------------------------------------#
# Test Class

class SyntaxExtRegTest(unittest.TestCase):

    def setUp(self):
        self.reg = synextreg.ExtensionRegister()

    def tearDown(self):
        self.reg.LoadDefault()

    #---- Test Cases ----#

    def testSingleton(self):
        self.assertTrue(self.reg is synextreg.ExtensionRegister())

    def testUnknownValue(self):
        val = self.reg['SomeFakeUnknownValue']
        self.assertTrue(val == u'txt')

    def testSetItem(self):
        self.assertRaises(TypeError, self.reg.__setitem__, 'test', 1)

    def testSerialize(self):
        ser = str(self.reg)
        self.assertTrue(isinstance(ser, basestring))
        for line in ser.split(os.linesep):
            self.assertTrue("=" in line)

    def testAssociate(self):
        self.reg.Associate(synextreg.LANG_TXT, "foo")
        ftype = self.reg.FileTypeFromExt("foo")
        self.assertEquals(ftype, synextreg.LANG_TXT)
        # An unknown extension should fall back to type Text
        ftype = self.reg.FileTypeFromExt("UNKNOWN_EXTENSION")
        self.assertEquals(ftype, synextreg.LANG_TXT)

    def testDisassociate(self):
        ftype = self.reg.FileTypeFromExt("pyw")
        self.assertEquals(ftype, synextreg.LANG_PYTHON)
        self.reg.Disassociate(synextreg.LANG_PYTHON, "pyw")
        ftype = self.reg.FileTypeFromExt("pyw")
        self.assertEquals(ftype, synextreg.LANG_TXT)

    def testFileTypeFromExt(self):
        ftype = self.reg.FileTypeFromExt("cpp")
        self.assertEquals(ftype, synextreg.LANG_CPP)
        ftype = self.reg.FileTypeFromExt("py")
        self.assertEquals(ftype, synextreg.LANG_PYTHON)
        ftype = self.reg.FileTypeFromExt("html")
        self.assertEquals(ftype, synextreg.LANG_HTML)

    def testGetAllExtensions(self):
        exts = self.reg.GetAllExtensions()
        self.assertTrue(isinstance(exts, list))
        for ext in exts:
            self.assertTrue(isinstance(ext, basestring), "Type Fail: %s" % ext)

    def testLoadDefault(self):
        self.reg.Associate(synextreg.LANG_CPP, "foo")
        ftype = self.reg.FileTypeFromExt("foo")
        self.assertTrue(ftype == synextreg.LANG_CPP, "FTYPE: %s" % ftype)
        self.reg.LoadDefault()
        ftype = self.reg.FileTypeFromExt("foo")
        self.assertTrue(ftype == synextreg.LANG_TXT)

    def testRemove(self):
        self.assertFalse(self.reg.Remove("UNKNOWN+TYPE"))
        self.assertTrue(self.reg.Remove(synextreg.LANG_4GL))

    def testSetAssociation(self):
        self.reg.Associate(synextreg.LANG_CPP, "foo")
        ftype = self.reg.FileTypeFromExt("foo")
        self.assertTrue(ftype == synextreg.LANG_CPP, "FTYPE: %s" % ftype)
        ftype = self.reg.FileTypeFromExt("cpp")
        self.assertTrue(ftype == synextreg.LANG_CPP, "FTYPE: %s" % ftype)

        # SetAssociation should replace existing associations with the
        # ones set by this call
        self.reg.SetAssociation(synextreg.LANG_CPP, "foo")
        ftype = self.reg.FileTypeFromExt("foo")
        self.assertTrue(ftype == synextreg.LANG_CPP)
        ftype = self.reg.FileTypeFromExt("cpp")
        self.assertTrue(ftype == synextreg.LANG_TXT, "FTYPE: %s" % ftype)

    #--- Module Level Function Tests ----#

    def testGetFileExtensionsFunct(self):
        exts1 = self.reg.GetAllExtensions()
        exts2 = synextreg.GetFileExtensions()
        self.assertTrue(exts1 == exts2)

    def testRegisterNewLangId(self):
        self.assertFalse("ID_LANG_FAKELANG" in dir(synextreg))
        nid = synextreg.RegisterNewLangId("ID_LANG_FAKELANG", "FakeLang")
        self.assertTrue(isinstance(nid, int))
        self.assertTrue("ID_LANG_FAKELANG" in dir(synextreg))




###############################################################################
# Name: ed_vim.py                                                             #
# Purpose: Vim emulation                                                      #
# Author: Hasan Aljudy                                                        #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Vim emulation class and helper methods to implement vim emulation in Editra's
text buffer.

"""

__author__ = "Hasan Aljudy"
__cvsid__ = "$Id: ed_vim.py 71184 2012-04-11 23:17:52Z CJP $"
__revision__ = "$Revision: 71184 $"

# ---------------------------------------------------------------------------- #
# Imports
import wx
import re
import string

# Local Imports
import ebmlib

# ---------------------------------------------------------------------------- #

# TODO:CJP maybe replace all repeatable methods with this decorator to reduce
#          repeated code. NOTE: may need to name different as the word repeat
#          is used in this module in many places.
#def repeat(funct, count=1):
#    def rwrap(*args, **kwargs):
#        for i in range(count):
#            funct()
#    return rwrap

# ---------------------------------------------------------------------------- #

def Parse(cmd, editor):
    """Parses a command and (if applicable) executes it on an editor
    @param cmd: the command string
    @param editor: a proxy object used to issue commands to the editor
    @return: boolean as a signal to the caller as to whether it should
             clear its command buffer.
             i.e. True if command is handled or invalid
             False if we need more input before we can execute the command

    """
    handled = True
    handler = GetHandler(cmd)
    if handler is None:
        if cmd.isdigit():
            handled = False
    else:
        need_more = DoHandle(handler, cmd, editor)
        if need_more:
            handled = False
    return handled

# ---------------------------------------------------------------------------- #

class EditraCommander(object):
    """Proxy object that sends commands to the editor

    The purpose of this class is to make vim support more abstract and portable
    so in theory one could add Vim Emulation to another editor using this module
    by just creating a commander object for it and re-implementing all of these
    methods defined below.

    The methods are based on turtle geometry. Everything is relative to the
    position of the cursor. Commands shouldn't take a position argument; they
    should all work on the current position of the cursor.

    """
    def __init__(self, keyprocessor):
        super(EditraCommander, self).__init__()

        # Attributes
        self.keyprocessor = keyprocessor
        self.stc = keyprocessor.stc
        self.LastChangeCommand = None
        self.LastInsertedText = None
        self.InsertRepeat = 1
        self.CaretStack = []
        self.ColumnStack = []
        self.ScrollStack = []
        self.LastFindChar = None
        self._Bookmarks = {}

    @property
    def STC(self):
        return self.stc

    def InsertMode(self):
        """Put editor in insert mode"""
        self.keyprocessor.InsertMode()

    def ReplaceMode(self):
        """Put editor in replace mode"""
        self.keyprocessor.ReplaceMode()

    def VisualMode(self):
        """Put editor in visual (selection) mode"""
        self.keyprocessor.VisualMode()

    def NormalMode(self):
        """Put editor in normal (command) mode"""
        self.keyprocessor.NormalMode()

    def IsNormalMode(self):
        return self.keyprocessor.IsNormalMode()

    def IsInsertMode(self):
        return self.keyprocessor.IsInsertMode()

    def InsertRepetition(self):
        """Repeat last inserted text according to the value that was set
        in SetInsertRepeat

        """
        if self.InsertRepeat > 1:
            self.RepeatChangeCommand(self.InsertRepeat - 1)
            self.InsertRepeat = 1

    def IsAtLineEnd(self):
        lnum = self.stc.GetCurrentLine()
        epos = self.stc.GetLineEndPosition(lnum)
        return epos == self._GetPos()

    def IsAtLineStart(self):
        """Is the cursor currently at the start of a line
        @return: bool

        """
        return self._GetCol() == 0

    def Undo(self, repeat):
        """Undo actions in the buffer
        @param repeat: int

        """
        for i in range(repeat):
            self.stc.Undo()

    def Redo(self, repeat):
        """Redo actions in the buffer
        @param repeat: int

        """
        for i in range(repeat):
            self.stc.Redo()

    def _GetPos(self):
        """Get caret position"""
        return self.stc.GetCurrentPos()

    def _SetPos(self, pos):
        """Set caret position"""
        self.stc.GotoPos(pos)

    def SelectLines(self, repeat):
        """Select specified number of lines starting with current line
        and going down
        @param repeat: int

        """
        cline = self.stc.GetCurrentLine()
        lline = self.stc.GetLineCount() - 1
        self.GotoLineStart()
        if cline == lline:
            cpos = self.stc.GetCurrentPos() - len(self.stc.GetEOLChar())
            cpos = max(0, cpos)
            self.stc.GotoPos(cpos)
        self.PushCaret()
        self.MoveDown(repeat)
        self.StartSelection()
        self.PopCaret()
        self.EndSelection()

    def PushCaret(self):
        """Pushes caret position for later restoration"""
        self.CaretStack += [self._GetPos()]

    def PopCaret(self, restore=True):
        """If no parameters are sent, restores caret position. If on the
        other hand, the restore keyword is set to false, the caret is not
        restored, but the last pushed value is discarded
        @keyword restore: bool

        """
        pos = self.CaretStack[-1]
        self.CaretStack = self.CaretStack[:-1]
        if restore:
            self._SetPos(pos)

    def PushColumn(self):
        """Push column position for later restoration"""
        self.ColumnStack += [self._GetCol()]

    def PopColumn(self, restore=True):
        """Pop caret position, and optionally discard it
        @keyword restore: if set to False, column will be discarded

        """
        column = self.ColumnStack[-1]
        self.ColumnStack = self.ColumnStack[:-1]
        if restore:
            self.GotoColumn(column)

    def PushScroll(self):
        """Push current scrolling state for later restoration"""
        self.ScrollStack += [self.stc.GetFirstVisibleLine()]

    def PopScroll(self, restore=True):
        """If not parameters are sent, restores the scrolling state. If on the
        other hadn, the restore keywords it set to false, then the last pushed
        scroll state is discarded
        @keyword restore: bool

        """
        line = self.ScrollStack[-1]
        self.ScrollStack[:-1]
        if restore:
            self.stc.ScrollToLine(line)

    def _IsLine(self, text):
        """Check if text ends with EOL
        @param text: string

        """
        eol = self.stc.GetEOLChar()
        eol_len = len(eol)
        if len(text) > eol_len:
            return text[-len(eol):] == eol
        else:
            return False

    def _GetCol(self):
        """Get the X position of the caret, aka column"""
        return self.stc.GetColumn(self._GetPos())

    def _SetCol(self, col):
        """Set the X position of the caret
        @param col: int

        """
        self.stc.GotoColumn(col)

    def GotoColumn(self, column):
        """Goto the specified column number on the current line"""
        self._SetCol(column)
        self.stc.ChooseCaretX()

    def StartSelection(self):
        """Record the starting place for selection
        Nothing is sent to the actual text control, just some internal
        state is recorded.
        @see: EndSelection

        """
        self._Anchor = self._GetPos()

    def EndSelection(self):
        """Set the selection to start from the starting position as it was set
        by StartSelection to the current caret position

        It doesn't matter whether the starting position is before or after the
        ending position

        """
        self.stc.SetSelection(self._Anchor, self._GetPos())

    def ExtendSelection(self):
        """If the selection was already ended, and the caret moved, this method
        will extend the selection to the new caret position.

        """
        self.EndSelection() # happens to be the same implementation ..

    def Deselect(self):
        """Unselect. Puts caret to the (visual) start of selection"""
        start, end = self._GetSelectionRange()
        self._SetPos(start)
        self.StartSelection()
        self.EndSelection()

    def GotoLine(self, line):
        """Goto the start of indentation of given line"""
        # PostionFromLine returns the position past the last character on the
        # line. So need to look at previous line to get the correct position.
        pos = self.stc.PositionFromLine(max(0,line-1))
        self._SetPos(pos)
        self.GotoIndentStart()

    def GotoLastLine(self):
        self.GotoLine(self.stc.GetLineCount())

    def GotoLineStart(self):
        """Goto the beginning of the current line"""
        self.stc.Home()

    def GotoIndentStart(self):
        """Goto to the first non-space character on the current line"""
        self.stc.GotoIndentPos()

    def GotoLineEnd(self):
        """Goto the end of the current line"""
        self.stc.LineEnd()

    def OpenLine(self):
        """Open a new line below the current line"""
        # TODO: watch out if AutoIndent gets "fixed" so that it doesn't
        #       open a new line by itself!
        self.GotoLineEnd()
        self.stc.AutoIndent()

    def OpenLineUp(self):
        """Open a new line above the current line"""
        self.MoveUp()
        self.OpenLine()

    def BeginUndoAction(self):
        self.stc.BeginUndoAction()

    def EndUndoAction(self):
        self.stc.EndUndoAction()

    def MoveUp(self, repeat=1):
        """Move caret up
        @keyword repeat: int

        """
        for i in range(repeat):
            self.stc.LineUp()

    def MoveDown(self, repeat=1):
        """Move caret down
        @keyword repeat: int

        """
        for i in range(repeat):
            self.stc.LineDown()

    def MoveRight(self, repeat=1):
        """Move position to the right
        @keyword repeat: int

        """
        self.stc.MoveCaretPos(repeat)

    def MoveLeft(self, repeat=1):
        """Move the caret postion to the left
        @keyword repeat: int

        """
        self.stc.MoveCaretPos(-repeat)

    def MoveForward(self, repeat=1):
        """Move the caret position to the right"""
        self.MoveRight(repeat)

    def MoveBack(self, repeat=1):
        """Move the caret position to the left"""
        self.MoveLeft(repeat)

    def NextWord(self, repeat=1):
        """Move to beginning of next word"""
        for i in range(repeat):
            self.stc.WordRight()

    def NextWordBig(self, repeat=1):
        """Move to beginning of next Word (words are separated by space)"""
        # TODO:CJP Test on empty document, possible error condition
        for i in range(repeat):
            self.stc.WordRight()
            while self.GetChar(-1) and not self.GetChar(-1).isspace():
                self.stc.WordRight()

    def WordEnd(self, repeat=1):
        """Move to end of this word"""
        for i in range(repeat):
            self.stc.WordRightEnd()

    def WordEndBig(self, repeat=1):
        """Move to end of this Word (words are separated by space)"""
        # TODO:CJP Test on empty document, possible error condition
        for i in range(repeat):
            self.stc.WordRightEnd()
            while self.GetChar() and not self.GetChar().isspace():
                self.stc.WordRightEnd()

    def BackWord(self, repeat=1):
        """Move back to start of word"""
        for i in range(repeat):
            self.stc.WordLeft()

    def BackWordBig(self, repeat=1):
        """Move back to start of Word (words are separated by space)"""
        # TODO:CJP Test on empty document, possible error condition
        for i in range(repeat):
            self.stc.WordLeft()
            while self.GetChar(-1) and not self.GetChar(-1).isspace():
                self.stc.WordLeft()

    def BackWordPart(self, repeat=1):
        """Move back to start of word part"""
        for i in range(repeat):
            self.stc.WordPartLeft()

    def WordPartEnd(self, repeat=1):
        """Move to the end of the word part"""
        for i in range(repeat):
            self.stc.WordPartRight()

    def ParaUp(self, repeat=1):
        """Move the caret one paragraph up"""
        for i in range(repeat):
            self.stc.ParaUp()

    def ParaDown(self, repeat=1):
        """Move the caret one paragraph down"""
        for i in range(repeat):
            self.stc.ParaDown()

    def RegexSearch(self, text, back=False):
        """Search for a regex expression
        @param text: search string
        @keyword back: seach in reverse

        """
        # TODO:CJP using the StyledTextCtrl's regex searching is not as
        #          powerful as using Editras SearchEngine.
        self.stc.SearchText(text, regex=True, back=back)
        if back and self.HasSelection():
            # Let the caret pos be at the end of the selection
            start, end = self._GetSelectionRange()
            self.stc.SetSelection(start, end)

    def GotoScreenTop(self):
        """Goto first visible line"""
        self.stc.GotoIndentPos(self.stc.GetFirstVisibleLine())

    def GotoScreenMiddle(self):
        """Goto to the middle line on the screen"""
        self.stc.GotoIndentPos(self.stc.GetMiddleVisibleLine())

    def GotoScreenLast(self):
        """Goto last visible line"""
        self.stc.GotoIndentPos(self.stc.GetLastVisibleLine())

    def _EnsureCaretVisible(self):
        """Smartly scroll view to make caret visible if it's out of view"""
        first = self.stc.GetFirstVisibleLine()
        lines = self.stc.LinesOnScreen()
        current = self.stc.GetCurrentLine()
        if current in range(first, first + lines):
            # supposedly visible! XXX: doesn't account for folding
            # TODO:CJP to ensure caret is visible call EnsureCaretVisible
            return None

        if current < first:
            amt = 0.7
        else:
            amt = 0.3
        self._Scroll(amt)

    def _Scroll(self, amt):
        """Scroll to the current line. For example _Scroll(0.5) scrolls the 
        current line to the middle of the screen
        @param amt: number between 0 and 1 signifying percentage

        """
        lines = self.stc.LinesOnScreen() - 1
        current = self.stc.GetCurrentLine()
        diff = int(lines * amt)
        self.stc.ScrollToLine(current - diff)

    def ScrollTop(self):
        """Scroll such that the current line is the first visible line"""
        self._Scroll(0)

    def ScrollMiddle(self):
        """Scroll such that the current line is in the middle of the screen"""
        self._Scroll(0.5)

    def ScrollBottom(self):
        """Scroll such that the current line is the last visible line"""
        self._Scroll(1)

    def _SelectIdentifierUnderCaret(self):
        """If nothing is selected, select the identifier under caret, otherwise
        keeps selection as is.
        @return: selected text

        """
        ident = self.GetSelectedText()
        if ident:
            return ident

        self.WordEnd()
        self.StartSelection()
        self.BackWord()
        self.EndSelection()
        ident = self.GetSelectedText()
        return ident

    def _NextIdent(self, repeat=1, back=False):
        """Jump to the next (or previous) occurance of identifier
        under caret or selected text.

        @note: Holds common code between NextIdent and PrevIdent
        @see: NextIdent
        @see: PrevIdent

        """
        ident = self._SelectIdentifierUnderCaret()
        self.PushCaret()
        self.PushScroll()
        if not ident:
            return

        for i in range(repeat):
            if back:
                self._GotoSelectionStart()
            else:
                self._GotoSelectionEnd()

            result = self.stc.SearchText(ident, back=back)
            if result == -1:
                break

        self.PopCaret(result == -1)
        self.PopScroll()
        self._EnsureCaretVisible()

    def NextIdent(self, repeat=1):
        """Find next occurance of identifier under cursor"""
        self._NextIdent(repeat)

    def PrevIdent(self, repeat=1):
        """Find the previous occurrence of identifier under cursor"""
        self._NextIdent(repeat, back=True)

    def InsertText(self, text):
        """Insert some text at the current caret position and advance caret
        The inserted text is remembered so it can be repeated.
        @param text: string

        """
        self.keyprocessor.InsertText(self._GetPos(), text)
        self.MoveForward(len(text))

    def SetInsertRepeat(self, repeat):
        """Set how many times to repeat the last inserted text when
        InsertRepetition is called.

        """
        self.InsertRepeat = repeat

    def SetLastChangeCommand(self, cmd):
        """Rememer this command as the last change cmmand so that it gets
        repeated when RepeatChangeCommand is called.

        """
        self.LastChangeCommand = cmd

    def SetLastInsertedText(self, text):
        """Set this as the last inserted text so it gets repeated when
        InsertRepetition is called.

        """
        self.LastInsertedText = text

    def RepeatChangeCommand(self, repeat=1):
        """Repeat the last change command as set by SetLastChangeCommand"""
        if not self.LastChangeCommand:
            return

        self.BeginUndoAction()
        for i in range(repeat):
            self.LastChangeCommand()
            self.InsertText(self.LastInsertedText)
        self.EndUndoAction()

    def GetChar(self, dist=1):
        """Get the character (or text) under caret
        @param dist: length of text to get, negative values can be used to get
            the text before the caret

        """
        pos = self._GetPos()
        start, end = minmax(pos, pos + dist)
        start = max(0, start)
        end = min(end, self.stc.GetLength())
        return self.stc.GetTextRange(start, end)

    def GetSelectedText(self):
        """Get the selected text
        @return: Unicode

        """
        return self.stc.GetSelectedText()

    def HasSelection(self):
        """Detects if there's anything selected
        @return: bool

        """
        return len(self.GetSelectedText()) > 0

    def SetRegister(self, reg='"'):
        """Set the current working clipboard (aka register) to given name"""
        ebmlib.Clipboard.Switch(reg)

    def YankSelection(self):
        """Copy the current selection to the clipboard"""
        ebmlib.Clipboard.Set(self.GetSelectedText())

    def DeleteSelection(self):
        """Yank selection and delete it"""
        start, end = self._GetSelectionRange()
        self.stc.BeginUndoAction()
        self.YankSelection()
        self.stc.Clear()
        self._SetPos(start)
        self.stc.EndUndoAction()

    def ChangeSelection(self):
        """Yank selection, delete it, and enter insert mode"""
        # HACK: need to adjust selection behavior for change command
        #       to better match vi behavior by not including trailing
        #       whitespace in the selection from the motion.
        stext = self.STC.GetSelectedText()
        clen = len(stext)
        slen = len(stext.rstrip())
        if slen < clen and slen > 0:
            start, end = self._GetSelectionRange()
            new_end = end - (clen - slen)
            self.STC.SetSelectionStart(start)
            self.STC.SetSelectionEnd(new_end)
        self.DeleteSelection()
        self.InsertMode()

    def Put(self, before=False, repeat=1):
        """Paste text

        If there's selection, the pasted text replaces it

        If text to be pasted ends with EOL it's assumed to be a line, and is
        pasted on a new line below the current line, or if the before keyword
        is set to True, then it's pasted on a new line above the current line

        """
        text = ebmlib.Clipboard.Get()
        if not text:
            return

        bIsLine = self._IsLine(text)
        if bIsLine:
            self.GotoLineStart()
            if not before:
                self.MoveDown()

        self.BeginUndoAction()
        if self.HasSelection():
            # paste over selection, if any
            self.stc.Clear()
        elif not bIsLine:
            self.stc.CharRight()

        for i in range(repeat):
            self.InsertText(text)
        self.EndUndoAction()

    def InvertCase(self, repeat):
        """Invert the case of the following characters"""
        if not self.HasSelection():
            for i in range(repeat):
                self.StartSelection()
                self.MoveForward()
                self.EndSelection()
        self.stc.InvertCase()
        self.Deselect()

    def JoinLines(self, repeat):
        """Join lines into a single line.
        @param repeat: number of lines below the current line to join with

        """
        self.SelectLines(repeat)
        self.stc.LinesJoinSelected()

    def _DeleteChars(self, repeat, back=False):
        """Delete characters under caret.
        @keyword back: If set to true, delete character before the caret

        """
        if back:
            move = self.MoveBack
        else:
            move = self.MoveForward

        self.BeginUndoAction()
        self.StartSelection()
        move(repeat)
        self.EndSelection()
        self.DeleteSelection()
        self.EndUndoAction()

    def DeleteForward(self, repeat):
        """Delete characters after the caret"""
        self._DeleteChars(repeat)

    def DeleteBackward(self, repeat):
        """Delete characters before the caret"""
        self._DeleteChars(repeat, back=True)

    def ReplaceChar(self, char, repeat):
        """Replace the next character under cursor with the given char"""
        self.BeginUndoAction()
        self.DeleteForward(repeat)
        for i in range(repeat):
            self.InsertText(char)
        self.EndUndoAction()

    def FindNextChar(self, char, repeat):
        """Find next occurance of char on the current line"""
        self.stc.FindChar(char, repeat)

    def FindPrevChar(self, char, repeat):
        """Find the previos occurance of char on the current line"""
        self.stc.FindChar(char, repeat, reverse=True)

    def FindTillNextChar(self, char, repeat):
        """Similar to FindNextChar, but stop one character short"""
        self.stc.FindChar(char, repeat, extra_offset=-1)

    def FindTillPrevChar(self, char, repeat):
        """Similar to FindPrevChar, but stop one character short"""
        self.stc.FindChar(char, repeat, reverse=True, extra_offset=1)

    def SetFindCharCmd(self, *args):
        """Remember last find-char command (raw unparsed command, e.g. 'f')
        for repeating later on.

        """
        self.LastFindChar = args

    def GetFindCharCmd(self):
        """@see SetFindCharCmd"""
        return self.LastFindChar

    def Mark(self, mark):
        """Create a bookmark on current line and give it a char label
        @param mark: the character label of the bookmark.

        """
        #TODO: handle global bookmarks
        bm_handle, column = self.stc.AddBookmark(), self._GetCol()
        if mark in string.ascii_lowercase:
             # Local bookmarks
            self._Bookmarks[mark] = (bm_handle, column)
        elif mark in string.ascii_uppercase:
            # Global bookmarks
            pass

    def GotoMark(self, mark):
        """Goto the position of the bookmark associated with the given
        character label.
        @param mark: the character label of the book mark
        @todo: hook into Bookmark Manager ed_bookmark?

        """
        if not mark in self._Bookmarks:
            return

        bm_handle, col = self._Bookmarks[mark]
        #TODO: handle global bookmarks
        line = self.stc.MarkerLineFromHandle(bm_handle)
        if line != -1:
            pos = self.stc.FindColumn(line, col)
            self.stc.GotoPos(pos)
        else: #line == -1 means the bookmark was deleted by the user
            del self._Bookmarks[mark]

    def _GetSelectionRange(self):
        """Get the range of selection such that the start is the visual start
        of the selection, not the logical start.

        """
        start, end = minmax(self.stc.GetSelectionStart(),
                            self.stc.GetSelectionEnd())
        return start, end

    def _GotoSelectionStart(self):
        """Goto the visual start of the current selection and deselect"""
        start, end = self._GetSelectionRange()
        self._SetPos(start)
        self.StartSelection()
        self.EndSelection()

    def _GotoSelectionEnd(self):
        """Goto the visual end of the current selection and deselect"""
        start, end = self._GetSelectionRange()
        self._SetPos(end)
        self.StartSelection()
        self.EndSelection()

    def _GetSelectedLines(self):
        """Get the first and last line (exclusive) of selection"""
        start, end = self._GetSelectionRange()
        start_line, end_line = (self.stc.LineFromPosition(start),
                                self.stc.LineFromPosition(end - 1) + 1)
        return start_line, end_line

    def SelectFullLines(self):
        """Extends selection so it covers entire lines"""
        start, end = self._GetSelectedLines()
        self.GotoLine(start)
        self.StartSelection()
        self.GotoLine(end)
        self.EndSelection()

    def IndentSelection(self, forward=True):
        """Indent the lines of the current selection forward (or backward)"""
        if not self.GetSelectedText():
            return

        self.PushScroll()
        self.PushCaret()
        self.BeginUndoAction()
        if forward:
            indent = self.stc.Tab
        else:
            indent = self.stc.BackTab

        for line in range(*self._GetSelectedLines()):
            self.stc.GotoLine(line)
            indent()
        self.EndUndoAction()
        self.PopCaret()
        self.PopScroll()

    def DedentSelection(self):
        """Indent lines of current selection backward"""
        self.IndentSelection(forward=False)

    def ShowCommandBar(self):
        """Open the command bar for the user to enter commands"""
        self.stc.ShowCommandBar()

    def ShowFindBar(self):
        """Open the find bar"""
        self.stc.ShowFindBar()

# ---------------------------------------------------------------------------- #

def minmax(a,b):
    return min(a,b), max(a,b)
# Internal functions, other modules should not bother with any of this

REPEAT_RE = re.compile("([1-9][0-9]*)*(.*)")

def SplitRepeat(cmd):
    """Split the command strings into a pair (repeat, rest)

    >>>SplitRepeat( '3ab' )
    (3, 'ab')
    >>>SplitRepeat( '13ab' )
    (13, 'ab')
    >>>SplitRepeat( 'abc' )
    (None, 'abc')
    >>>SplitRepeat( 'ab2' )
    (None, 'ab2')
    >>>SplitRepeat( '0' )
    (None, '0')
    >>>SplitRepeat( '0ab' )
    (None, '0ab')
    >>>SplitRepeat('8')
    (8, '')

    """
    repeat, rest = re.match( REPEAT_RE, cmd ).groups()
    if repeat:
        repeat = int(repeat)
    return (repeat, rest)

def GetHandler(cmd, h_list=None):
    """Finds the function that handles command cmd
    @param cmd: the command string
    @keyword h_list: alternative list of handler functions

    """
    if h_list is None:
        h_list = HANDLERS

    repeat, cmd = SplitRepeat(cmd)
    if not cmd:
        return None

    for handler in h_list:
        if cmd[0] in handler.start_chars:
            return handler

def DoHandle(handler, cmd, editor):
    """Call handler for command
    It's necessary to use this function instead of calling the handler
    directly.

    """
    repeat, cmd = SplitRepeat(cmd)
    if handler.generic_repeat:   # generic_repeat means that if no repeat
        repeat = repeat or 1     # parameter is supplied, use 1
    return handler(editor, repeat, cmd)

def GetMotion(editor, cmd):
    """Move cursor of editor to a new location according to motion
    Returns a method that handles this motion, or None

    """
    if not cmd:
        return None

    handler = GetHandler(cmd, MOTION_HANDLERS)
    if handler is None:
        return None

    def motion_function():
        return DoHandle(handler, cmd, editor)
    return motion_function

# ---------------------------------------------------------------------------- #
# Vim commands

def vim_parser(start_chars, generic_repeat=True, is_motion=False):
    """Decorator for function that handle vim commands

    Command Handling functions must follow the signature:

        function( editor, repeat, cmd )

    These functions should never be called directly, but instead
    be called through the GetHandler and DoHandle functions.

    The handler functions should always decorate themselves with this
    decorator.

    The purpose of this decorator is to indicate:
        * What is the first char of commands that the handler can handle
            - So that GetHandler can find the correct handler for a vim command
        * Whether it's a motion
            - Some commands operate on motions
        * How to handle the repeat parameter
            - For most command, no repeat is the same as 1 repeat, but for
              some commands, this is not the case

    """
    def decorator(handler):
        handler.is_parser = True
        handler.start_chars = start_chars
        handler.generic_repeat = generic_repeat
        handler.is_motion = is_motion
        return handler
    return decorator

# TODO replace with a special objects
# TODO:CJP ok if these become classes but, if they remain as flags then they
#          need to be uppercase.
NeedMore = True
InvalidOp = False

@vim_parser( u'iIaAoO' )
def InsertMode(editor, repeat, cmd):
    """Handler for basic commands that put vim in INSERT mode"""
    cmd_map = {
            u'i': (lambda:None),
            u'I': editor.GotoIndentStart,
            u'A': editor.GotoLineEnd,
            u'a': editor.MoveForward,
            u'o': editor.OpenLine,
            u'O': editor.OpenLineUp
            }
    func = cmd_map[cmd]
    func()
    editor.SetLastChangeCommand(func)  # Let the editor remember to repeat
    editor.SetInsertRepeat(repeat)     # insertion if and when needed
    editor.InsertMode()

#TODO:CJP Documentation for below methods please
@vim_parser( u'v' )
def VisualMode(editor, repeat, cmd):
    """Enter visual mode
    @see: vim_parser

    """
    editor.VisualMode()

@vim_parser('.')
def Dot(editor, repeat, cmd):
    """Repeat last insert command
    @see: vim_parser

    """
    editor.RepeatChangeCommand(repeat)
    if editor.IsInsertMode():
        editor.NormalMode() # in case it was a 'c' command

@vim_parser(u'hjkl\r \x08\u013c\u013a', is_motion=True)
def Arrows(editor, repeat, cmd):
    """Basic arrow movement in vim.
    @see: vim_parser

    """
    cmd_map = {
            u'h': editor.MoveLeft,
            u'j': editor.MoveDown,
            u'k': editor.MoveUp,
            u'l': editor.MoveRight,
            u'\r': editor.MoveDown,
            u' ' : editor.MoveRight,
            u'\x08' : editor.MoveLeft
          }
    if cmd in cmd_map:
        cmd_map[cmd](repeat)
    else:
        # Handle motion for actual arrow keys
        if cmd == u'\u013c' and not editor.IsAtLineEnd():
            # Right arrow
            editor.MoveRight()
        elif cmd == u'\u013a' and not editor.IsAtLineStart():
            # Left Arrow
            editor.MoveLeft()

@vim_parser('wbeWBE[]', is_motion=True)
def Words(editor,repeat, cmd):
    """Word motions.
    @note: [] is based on what scintilla calls "word parts", these motions
           are not in vim
    @see: vim_parser

    """
    cmd_map = {
            u'w': editor.NextWord,
            u'e': editor.WordEnd,
            u'b': editor.BackWord,
            u'W': editor.NextWordBig,
            u'E': editor.WordEndBig,
            u'B': editor.BackWordBig,
            u'[': editor.BackWordPart,
            u']': editor.WordPartEnd,
          }
    cmd_map.get(cmd, lambda x: None)(repeat)

@vim_parser(u'$^0', is_motion=True)
def Line(editor, repeat, cmd):
    """Motions to beginning/end of a line.
    @see: vim_parser

    """
    cmd_map = { u'0': editor.GotoLineStart,
                u'^': editor.GotoIndentStart,
                u'$': editor.GotoLineEnd,
              }
    cmd_map.get(cmd, lambda:None)()

@vim_parser(u'{}', is_motion=True)
def Para(editor, repeat, cmd):
    """Paragraph motions.
    @see: vim_parser

    """
    cmd_map = { u'{': editor.ParaUp,
            u'}': editor.ParaDown,
          }
    cmd_map.get(cmd, lambda x: None)(repeat)

@vim_parser(u'uU')
def Undo(editor, repeat, cmd):
    """Undo/Redo commands.
    @note: unlike vim, U is used for redo.
    @see: vim_parser

    """
    if cmd == u'u':
        editor.Undo(repeat)
    elif cmd == u'U':
        editor.Redo(repeat)

# TODO:CJP Scintilla lexing bug, check if fixed in 2.9
@vim_parser(u'*#', is_motion=True)
def FindIdent(editor, repeat, cmd):
    """Find the next/previous occurance of identifier under caret
    @note: There's a sublte difference from vim: if some text is already
           selected, it's used as the search term instead of the identifier
           under the caret
    @see: vim_parser

    """
    cmd_map = { u'#': editor.PrevIdent,
                u'*': editor.NextIdent,
              }
    cmd_map.get(cmd, lambda x: None)(repeat)

@vim_parser(u'~')
def Tilde(editor, repeat, cmd):
    """Invert case of character(s) under caret. Also operates on selection.
    @see: vim_parser

    """
    editor.InvertCase(repeat)
    editor.NormalMode() # In case this command was applied in visual mode

@vim_parser(u'J')
def Join(editor, repeat, cmd):
    """Join lines command.
    @see: vim_parser

    """
    editor.PushCaret()
    editor.JoinLines(repeat)
    editor.PopCaret()

@vim_parser(u'xXsSCDY')
def Delete(editor, repeat, cmd):
    """Simple delete/change commands that are implemented in terms of more
    advanced c/d commands.
    @see: vim_parser

    """
    cmd_map = {
        u'x': u'dl', u'X': u'dh',
        u's': u'cl', u'S': u'cc',
        u'C': u'c$', u'D': u'd$', u'Y': u'y$',
        }
    if cmd in cmd_map:
        Change(editor, repeat, cmd_map[cmd])

@vim_parser(u'cdy<>')
def Change(editor, repeat, cmd):
    """Implementations for c/d/y commands. Also for <> (indentation) commands.
    @todo: This method is quite larger than other methods in this module,
           needs to be simplified.
    @see: vim_parser

    """
    editor.PushColumn()
    editor.PushCaret()
    def ret(return_value, restore=True):
        """Needed because we have to pop what we pushed before we return"""
        editor.PopCaret(restore)
        editor.PopColumn(restore)
        return return_value
    pre_selected = False
    line_motion = False
    if editor.HasSelection():
        pre_selected = True
        if len(cmd) > 1:
            motion = cmd[1:]
        else:
            motion = ''
        cmd = cmd[0]
    else:
        if len(cmd) == 1:
            return ret(NeedMore, False)

        cmd, motion = cmd[0], cmd[1:]
        if motion.isdigit():
            return ret(NeedMore)

        motion_repeat, motion_cmd = SplitRepeat(motion)
        if motion_repeat:
            repeat = repeat * motion_repeat

        if motion_cmd == cmd:
            # Operate on whole line
            line_motion = True
            editor.PushColumn()
            editor.SelectLines(repeat)
        else:
            motion_function = GetMotion(editor, motion_cmd)
            if motion_function is None:
                # Invalid motion; cancel operation
                return ret(InvalidOp)

            editor.PushCaret()
            editor.PushColumn()
            for i in range(repeat):
                if motion_function() == NeedMore:
                    # This motion is incomplete .. (catch on first iteration)
                    editor.Deselect()
                    return ret(NeedMore)
            editor.StartSelection()
            editor.PopCaret()
            editor.EndSelection()

            if motion_cmd[0] in LINE_MOTION_PREFIXES:
                line_motion = True
                editor.SelectFullLines()

    cmd_map = {
            u'y': editor.YankSelection,
            u'd': editor.DeleteSelection,
            u'c': editor.ChangeSelection,
            u'<': editor.DedentSelection,
            u'>': editor.IndentSelection,
          }

    cmd_map[cmd]()

    restore_x = cmd in (u'y', u'<', u'>') or cmd == u'd' and line_motion
    editor.PopColumn(restore_x)

    # XXX: Some special case handling
    #      Not the most elegant way though ..
    if cmd == u'c':
        pass
    else:
        # Repeating delete/yank/indent commands doesn't insert any text
        editor.SetLastInsertedText(u'')
        # Applying them in visual mode ends visual mode
        editor.NormalMode()

    if not pre_selected:
        # Remember this command as last change
        # However, if we're operating on a selection, then remembering
        # doesn't make much sense
        editor.SetLastChangeCommand(lambda : Change(editor, repeat, cmd+motion))

@vim_parser(u'pP')
def Put(editor, repeat, cmd):
    """Paste commands.
    @see: vim_parser

    """
    before = cmd == u'P'
    editor.Put(before, repeat)
    editor.NormalMode() # for pasting in visual mode

@vim_parser(u'"')
def Reg(editor, repeat, cmd):
    """Switch register (clipboard) command
    @note: This command is stand-alone, not prefix to other commands.
    @see: vim_parser

    """
    if len(cmd) == 1:
        return NeedMore

    cmd, char = cmd
    try:
        editor.SetRegister(char)
    except ebmlib.ClipboardException:
        # Attempted to switch to invalid register
        wx.Bell()

@vim_parser(u'r')
def ReplaceChar(editor, repeat, cmd):
    """Replace character under caret with another one.
    @note: Does not enter into Insert mode.
    @see: vim_parser

    """
    if len(cmd) == 1:
        return NeedMore

    r, char = cmd
    f = lambda: editor.ReplaceChar(char, repeat)
    f()
    editor.SetLastChangeCommand(f)
    editor.SetLastInsertedText(u'')

@vim_parser(u'R')
def ReplaceMode(editor, repeat, cmd):
    """Enter into Replace Mode.
    @see: vim_parser

    """
    # TODO handle repetition
    editor.ReplaceMode()

@vim_parser(u'fFtT', is_motion=True)
def FindChar(editor, repeat, cmd):
    """Find character on current line and move the caret to it (if found).
    @see: vim_parser

    """
    if len(cmd) == 1:
        return NeedMore

    cmd, char = cmd
    cmd_map = {
            u'f' : editor.FindNextChar,
            u'F' : editor.FindPrevChar,
            u't' : editor.FindTillNextChar,
            u'T' : editor.FindTillPrevChar,
          }
    if cmd in cmd_map:
        cmd_map[cmd](char, repeat)
        editor.SetFindCharCmd(cmd, char)

@vim_parser(u',;', is_motion=True)
def RepeatFindChar(editor, repeat, cmd):
    """Repeat the last FindChar motion.
    @see: vim_parser

    """
    prev_cmd = editor.GetFindCharCmd()
    if not prev_cmd:
        return

    pcmd, char = prev_cmd
    if cmd == u',':
        pcmd = pcmd.swapcase()

    FindChar(editor, repeat, pcmd+char)
    editor.SetFindCharCmd(*prev_cmd)

@vim_parser(u'm')
def Mark(editor, repeat, cmd):
    """Create a bookmark and associate it with a character label.
    @see: vim_parser

    """
    if len(cmd) == 1:
        return NeedMore

    cmd, mark = cmd
    editor.Mark(mark)

@vim_parser(u"`'", is_motion=True)
def Jump(editor, repeat, cmd):
    """Jump to a bookmark specified by a character label.
    @see: vim_parser

    """
    if len(cmd) == 1:
        return NeedMore

    cmd, mark = cmd
    editor.GotoMark(mark)
    if cmd == u"'":
        editor.GotoIndentStart()

@vim_parser(u'HLMGgz', is_motion=True, generic_repeat=False)
def NavExtra(editor, repeat, cmd):
    """Commands for navigating visible lines or scrolling.
    @see: vim_parser

    """
    if cmd == u'H':
        editor.GotoScreenTop()
        if repeat:
            editor.MoveDown(repeat)
    elif cmd == u'M':
        editor.GotoScreenMiddle()
    elif cmd == u'L':
        editor.GotoScreenLast()
        if repeat:
            editor.MoveUp(repeat)
    elif cmd == u'G':
        if repeat:
            editor.GotoLine(repeat)
        else:
            editor.GotoLastLine()
    else: # g or z command
        if len(cmd) < 2:
            return NeedMore

        if cmd == u'gg':
            editor.GotoLine(0)
        elif cmd == u'zz':
            editor.ScrollMiddle()
        elif cmd == u'zt':
            editor.ScrollTop()
        elif cmd == u'zb':
            editor.ScrollBottom()

@vim_parser(u'/?', is_motion=True)
def RegexSearch(editor, repeat, cmd):
    """Incremental search commands.
    @note: Uses the find bar in editra.
    @see: vim_parser

    """
    # XXX: findbar is not regex!
    # TODO:CJP it is if you set the regex search flag or the user's previous
    #          search was done using regular expressions.
    editor.ShowFindBar()

@vim_parser(u'|', is_motion=True)
def Column(editor, repeat, cmd):
    """Goto specified column, with 1 being the default.
    @see: vim_parser

    """
    editor.GotoColumn(repeat)

@vim_parser(u':')
def Ex(editor, repeat, cmd):
    """Command for opening the command bar.
    @see: vim_parser

    """
    editor.ShowCommandBar()

# Cannonical list of handlers
HANDLERS = (
    InsertMode, ReplaceMode, VisualMode, Dot, Undo,
    Arrows, Words, Line, Para, Column, FindChar, RepeatFindChar,
    FindIdent, Change, Delete, Tilde, ReplaceChar, Put, Join,
    Mark, Jump, Ex, Reg, NavExtra, RegexSearch,
    )
MOTION_HANDLERS = [h for h in HANDLERS if h.is_motion]
# HACK: The following is a sign of improper design! but doing it properly
#       would probably require yet another huge rewrite of this module.
#       So watch out if things get out of control with many such hacks.
# These are motions that, if operated on, the operation should happen
# on whole lines.
LINE_MOTION_PREFIXES = [u'\'', u'G', u'{', u'}', u'H', u'L', u'M']



###############################################################################
# Name: ooc.py                                                                #
# Purpose: Define OOC Syntax Highlighting Support and other features          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2010 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: ooc.py                                                                
@author: Cody Precord                                                       
@summary: Lexer configuration file for OOC (Out of Class)
                                                                         
"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id:  $"
__revision__ = "$Revision:  $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local imports
import synglob
import syndata
import _cpp

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

KEYWORDS = (0, "class cover interface implement func abstract extends from "
               "this super new const final static include import use extern "
               "inline proto break continue fallthrough operator if else for "
               "while do switch case as in version return true false null "
               "default")

TYPES = (1, "Int Int8 Int16 Int32 Int64 Int80 Int128 UInt UInt8 UInt16 UInt32 "
            "UInt64 UInt80 UInt128 Octet Short UShort Long ULong LLong ULLong "
            "Float Double LDouble Float32 Float64 Float128 Char UChar WChar "
            "String Void Pointer Bool SizeT This")

#------------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for many OOC""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_CPP)

    def GetKeywords(self):
        """Returns Specified Keywords List"""
        return [KEYWORDS, TYPES, _cpp.DOC_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return _cpp.SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set"""
        return [_cpp.FOLD, _cpp.FOLD_COM]

    def GetCommentPattern(self):
        """Return comment pattern for OOC"""
        return [u'//']



###############################################################################
# Name: FileTreeDemo.py                                                       #
# Purpose: Test and demo file for eclib.FileTree                              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2011 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Test file for testing the FileTree control

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id:  $"
__revision__ = "$Revision:  $"

#-----------------------------------------------------------------------------#
# Imports
import os
import sys
import wx

import IconFile

# Put local package on the path
#sys.path.insert(0, os.path.abspath('../../src'))
import eclib

#-----------------------------------------------------------------------------#

class TestPanel(wx.Panel):
    def __init__(self, parent, log):
        super(TestPanel, self).__init__(parent)

        # Attributes
        self.ftree = eclib.FileTree(self)
        self.ftree.AddWatchDirectory(wx.GetUserHome())
        self.selbtn = wx.Button(self, label="Selected Files")
        self.addbtn = wx.DirPickerCtrl(self, message="Add Watch",
                                       style=wx.DIRP_CHANGE_DIR)
        self.addbtn.PickerCtrl.SetLabel("Add Watch")
        self.rmbtn = wx.Button(self, label="Remove Watch")
        self.setselbtn = wx.Button(self, label="Select a File")
        self.log = log

        # Layout
        self.__DoLayout()

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnGetSelected, self.selbtn)
        self.Bind(wx.EVT_DIRPICKER_CHANGED, self.OnAddWatch, self.addbtn)
        self.Bind(wx.EVT_BUTTON, self.OnRemoveWatch, self.rmbtn)
        self.Bind(wx.EVT_BUTTON, self.OnSelectFile, self.setselbtn)

    def __DoLayout(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.ftree, 1, wx.EXPAND)
        btnsz = wx.BoxSizer(wx.HORIZONTAL)
        btnsz.Add(self.addbtn)
        btnsz.Add(self.rmbtn)
        btnsz.Add(self.selbtn)
        btnsz.Add(self.setselbtn)
        sizer.Add(btnsz)
        self.SetSizer(sizer)

    def OnAddWatch(self, evt):
        val = self.addbtn.GetPath()
        self.log.write("Add new watch: %s" % val)
        self.ftree.AddWatchDirectory(val)

    def OnRemoveWatch(self, evt):
        ch = wx.GetSingleChoice("Select watch dir to remove", "Remove Watch",
                                self.ftree.WatchDirs, self)
        if ch:
            self.ftree.RemoveWatchDirectory(ch)

    def OnGetSelected(self, evt):
        sel = self.ftree.GetSelectedFiles()
        self.log.write(repr(sel))

    def OnSelectFile(self, evt):
        ddir = ""
        if len(self.ftree.WatchDirs):
            ddir = self.ftree.WatchDirs[0]
        dlg = wx.FileDialog(self, "Select a file", ddir,
                            style=wx.FD_OPEN)
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
            self.log.write("SelectFile(%s)" % fname)
            self.ftree.SelectFile(fname)
        dlg.Destroy()

#-----------------------------------------------------------------------------#

def runTest(frame, nb, log):
    win = TestPanel(nb, log)
    return win

class TestLog:
    def __init__(self):
        pass

    def write(self, msg):
        print msg

#----------------------------------------------------------------------

overview = eclib.FileTree.__doc__
title = "FileTree"

#-----------------------------------------------------------------------------#
if __name__ == '__main__':
    try:
        import run
    except ImportError:
        app = wx.App(False)
        frame = wx.Frame(None, title="FileTree Demo")
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(TestPanel(frame, TestLog()), 1, wx.EXPAND)
        frame.CreateStatusBar()
        frame.SetSizer(sizer)
        frame.SetInitialSize((300, 300))
        frame.Show()
        app.MainLoop()
    else:
        run.main(['', os.path.basename(sys.argv[0])] + sys.argv[1:])
